#!/usr/bin/python2.7

"""
Additional HTTP API for Python Package Index based on DevPI
"""

from .API import HttpAPI;
import re;
import xml.etree.ElementTree as ET;
import posixpath;

class DevPIAPI( HttpAPI ):
    """
    Class for providing some simple Http requests to PyPI based on DevPI 
    """
    _env_prefix = "PYPI_PRODUCTION";

    def _split_pkg_string( r_self, str_name ):
        """
        Split string for package name into components following PyPI standard
        :param r_self: self reference
        :type r_self: DevPIApi object
        :param str_name: python package name
        :type str_name: string, unicode
        :return: dictionary with components: distribution, version, build_tag, python_tag, abi_tag, platform_tag. Some of them may be absent
        """

        obj_match = re.match( r'^(?P<distribution>[^\.]*)-(?P<version>[\d\.]+)(-(?P<build_tag>[^\-]+))?-(?P<python_tag>[^\-]+)-(?P<abi_tag>[^\-]+)-(?P<platform_tag>[^\-\.]+)\.whl$', str_name );

        if obj_match is None:
            obj_match = re.match( r'^(?P<distribution>[^\.]*)-(?P<version>[\d\.]+)\.tar\.gz$', str_name );

        if obj_match is None:
            raise ValueError( 'DevPI answer is wrong: package name ' + str_name + " does not match any known pattern" );

        return obj_match.groupdict();

    def get_package_versions( r_self, str_name ):
        """
        Get package version list from server
        :param r_self: self reference
        :type r_self: DevPIApi object
        :param str_name: python package name
        :type str_name: string, unicode
        :return: list of versions on the server, or None if nothing found
        """

        str_name = str_name.strip();
        str_request = posixpath.join( "+simple", str_name );
        obj_resp = r_self.get( str_request );

        if ( obj_resp.status_code != 200 ):
            return None;

        xml_resp = ET.fromstring( obj_resp.text );

        return list( set( [r_self._split_pkg_string( x.text )[ 'version' ] for x in xml_resp.findall( 'body/a' )] ) )



import json, logging, os, re, time

from oc_cdtapi import API
import posixpath

class DmsGetverAPI (API.HttpAPI):
    # prefix for credentials environment variables used by HttpAPI
    _env_prefix = 'DMS'


    def __init__ (self, *args, **argp):
        logging.debug ('Reached __init__')
        logging.debug ('Calling base class constructor for availability of HttpAPI methods')
        super (DmsGetverAPI, self).__init__ (*args, **argp)

        # delivery states in process
        self.waiting_states = ['INITIATED', 'PROCESSING', 'QUEUED']

        # delivery states finished
        self.exit_states = ['READY', 'FAILED']

        # wait for state timeout
        self.wait_state_timeout = 1000

        # wait for state request interval
        self.wait_state_sleep = 30


    def create_distr_request (self, version = None, source_version = None, distr_type = None, client_filter = None):
        """
        Creates a new distribution request
        :parameter version: required version e.g. 03.44.30.55
        :parameter distr_type: distribution type e.g. CARDS
        :parameter client_filter: a set of software components e.g. Diners Club Russia Acquiring;MasterCard;VISA;
                                  if not provided an attempt to fetch it from svn will be performed
                                  filter fetching is defined in separate class
        :return: distribution state info as returned by dms
        """
        logging.debug ('Reached create_distr_request')

        distr_state_info = self._create_distr_request_int (version = version, source_version = source_version, distr_type = distr_type, client_filter = client_filter)
        return distr_state_info


    def get_distr (self, distr_id, distr_option):
        """
        Fetches distribution from dms
        :parameter distr_id: distribution id
        :return distr: distribution fetched from dms or None on error
        :return distr_state_info: distribution info
        """
        logging.debug ('Reached get_distr')
        logging.debug ('Distr id [%s]' % distr_id)
        logging.debug ('Distr option [%s]' % distr_option)
        distr = None
        distr_state_info = self.get_distr_state_info_byid (distr_id, distr_option)
        if distr_state_info ['state'] != 'READY':
            logging.debug ('Distribution [%s] is in not ready state [%s]' % (distr_id, distr_state_info ['state'] ) )
        else:
            logging.debug ('Distribution [%s] is in ready state, downloading' % distr_id)
            url = posixpath.join('dms-getver', 'distribution', 'id:%s' % distr_id, 'download')
            logging.debug ('Getting from [%s]' % url)
            logging.debug ('Requesting [%s]' % distr_state_info ['fileName'] )
            distr_resp = self.get (url)
            if distr_resp.status_code != 200:
                logging.debug ('DMS returned an error response [%s] while getting [%s]' % (distr_resp.status_code, url) )
            else:
                distr = distr_resp.content
                logging.debug ('Fetched [%s] bytes from [%s]' % (len (distr), url) )
        return distr, distr_state_info


    def get_distr_state_info (self,  version = None, source_version = None, distr_type = None, client_filter = None):
        """
        Requests distribution state info from dms
        parameters description see in create_distr_request
        """
        logging.debug ('Reached get_distr_state_info')

        url, parms = self._get_distr_state_url (version = version, source_version = source_version, distr_type = distr_type, client_filter = client_filter)
        distr_state_info = self._get_distr_state_info_int (url, parms)
        return distr_state_info


    def get_distr_state_info_byid (self, distr_id, distr_option):
        """
        Requests distribution state info from dms
        """
        logging.debug ('Reached get_distr_state_info_byid')
        if distr_option == 'full':
            logging.debug ('Full distr info requested')
            url = posixpath.join('dms-getver', 'rest', 'api', '1', 'distribution', 'id:%s' % distr_id)
        else:
            logging.debug ('Diff distr info requested')
            url = posixpath.join('dms-getver', 'rest', 'api', '1', 'distribution-difference', 'id:%s' % distr_id)

        distr_state_info = self._get_distr_state_info_int (url)
        return distr_state_info


    def get_dms_gav (self, distr_id, distr_option):
        """
        Requests gav by distribution id
        """
        logging.debug ('Reached get_dms_gav')
        logging.debug ('Request for gav for [%s] distr id [%s]' % (distr_option, distr_id) )
        if distr_option == 'full':
            url = posixpath.join('dms-getver', 'rest', 'api', '1', 'distribution', 'id:%s' % distr_id, 'gav')
        else:
            url = posixpath.join('dms-getver', 'rest', 'api', '1', 'distribution-difference', 'id:%s' % distr_id, 'gav')
        resp = self.get (url)
        if resp.status_code == 200:
            logging.debug ('OK response from dms-getver')
            gav = resp.json ()
            gav_text = '%s:%s:%s:%s' % (gav ['groupId'], gav ['artifactId'], gav ['version'], gav ['packaging'] )
            logging.debug ('Returning [%s]' % gav_text)
        else:
            logging.debug ('Error response [%s] from dms-getver' % resp.status_code)
            gav_text = None
            logging.debug ('Returning None')
        return gav_text


    def get_dms_log (self, distr_id, distr_option):
        """
        Retrieve dms log
        """
        logging.debug ('Reached get_dms_log')
        logging.debug ('Request for log of processing distr [%s]' % distr_id)

        url = self.get_dms_log_url (distr_id, distr_option)

        log = None
        resp = self.get (url)
        status_code = resp.status_code
        logging.debug ('Response status code: [%s]' % status_code)

        if status_code != 200:
            logging.debug ('Error response from dms')
        else:
            log = resp.text

        return log


    def get_dms_log_url (self, distr_id, distr_option):
        logging.debug ('Reached get_dms_log_url')
        logging.debug ('Request for url for [%s] distr [%s]' % (distr_option, distr_id) )
        if distr_option == 'full':
            url = posixpath.join('dms-getver', 'rest', 'api', '1', 'distribution', 'id:%s' % distr_id, 'log')
        else:
            url = posixpath.join('dms-getver', 'rest', 'api', 'distribution-difference', 'id:%s' % distr_id, 'log')
        logging.debug ('url for log request: [%s]' % url)
        return url


    def wait_for_state (self, distr_id, distr_option):
        """
        Wait until either distr gets into an exit state or a timeout occurs
        States and timeouts are defined in __init__
        """
        logging.debug ('Reached wait_for_state')
        ela = 0
        st = int (time.time () )
        distr_state_info = {}
        distr_state_info ['state'] = 'TIMEOUT'
        while ela < int (self.wait_state_timeout):
            ela = int (time.time () ) - st
            dsi = self.get_distr_state_info_byid (distr_id, distr_option)
            state = dsi ['state']
            if not state in self.waiting_states:
                logging.debug ('Distr [%s] is in exit state [%s]' % (distr_id, state) )
                distr_state_info = dsi
                break
            logging.debug ('Distr [%s] is in waiting state [%s]' % (distr_id, state) )
            logging.debug ('Retrying in [%s] sec., [%s] of [%s] sec. in wait' % (self.wait_state_sleep, ela, self.wait_state_timeout) )
            time.sleep (self.wait_state_sleep)
        return distr_state_info


    def _create_distr_request_int (self, version = None, source_version = None, distr_type = None, client_filter = None):
        """
        Creates a new distribution request
        """
        logging.debug ('Reached _create_distr_request_int')
        logging.debug ('version = [%s]' % version)
        logging.debug ('source_version = [%s]' % source_version)
        logging.debug ('distr_type = [%s]' % distr_type)
        if client_filter:
            logging.debug ('client_filter length = [%s]' % len (client_filter) )
        else:
            logging.debug ('Filters not specified')
            client_filter = ''

        distr_state_info = {}
        req_parm = {}
        req_parm ['product'] = distr_type

        if source_version:
            logging.debug ('Diff distribution requested')
            url = posixpath.join('dms-getver', 'rest', 'api', '1', 'distribution-difference')
            req_parm ['initialVersion'] = source_version
            req_parm ['targetVersion'] = version
            req_parm ['initialFilters'] = client_filter
            req_parm ['targetFilters'] = client_filter
        else:
            logging.debug ('Full distribution requested')
            url = posixpath.join('dms-getver', 'rest', 'api', '1', 'distribution')
            req_parm ['version'] = version
            req_parm ['filter'] = client_filter

        logging.debug ('request parameters: [%s]' % json.dumps (req_parm, indent = 4) )
        resp = self.post (url, json = req_parm)

        # we shold have been expecting 202 here, but it's 200 somehow
        if resp.status_code != 200:
            logging.debug ('DMS responds with unexpected status code [%s]' % resp.status_code)
            logging.debug ('DMS response body: %s' % resp.text)
            distr_state_info ['id'] = None
            distr_state_info ['state'] = 'HTTP/%s' % resp.status_code
        else:
            distr_state_info = resp.json ()

        distr_state = distr_state_info ['state']
        distr_id = distr_state_info ['id']

        if distr_state and distr_id:
            logging.debug ('Created dist id [%s] in state [%s]' % (distr_id, distr_state) )
            distr_state_info = self._normalize_dsi (distr_state_info)

        return distr_state_info


    def _dumb_404 (self, resp):
        """
        This is to distinguish a server 404 from an application 404 (which should have been 200)
        """
        logging.debug ('Reached _dumb_404')
        try:
            j = resp.json ()
        except ValueError as e:
            logging.debug ('Failed to get json from dms response, returning False')
            return False
        if j ['code'] == 'DMS-GETVER-40001':
            logging.debug ('Found a "dms not found" response code, returning True')
            return True
        else:
            logging.debug ('Json response processed, but no "dms not found" code detected, returning False')
            return False


    def _get_distr_state_info_int (self, url, parms=None):
        """
        Requests and returns distr state
        states known so far: INITIATED, PROCESSING, FAILED, READY. + 
        NOTFOUND set by this method upon http/404 and TIMEOUT set by wait_for_state
        """
        logging.debug ('Reached _get_distr_state_info_int')
        logging.debug ('URL = [%s]' % url)

        try:
            resp = self.get (url, params=parms)
        except API.HttpAPIError as e:
            resp = e.resp

        distr_state_info = {}

        # bad status
        if resp.status_code != 200:
            logging.debug ('DMS reponds with status code [%s]' % resp.status_code)
            if resp.status_code == 404 and self._dumb_404 (resp):
                distr_state_info ['state'] = 'NOTFOUND'
                distr_state_info ['id'] = None
            else:
                distr_state_info ['state'] = 'HTTP/%s' % resp.status_code
                distr_state_info ['id'] = None
        # ok status
        else:
            distr_state_info = resp.json ()

        distr_id = distr_state_info ['id']
        distr_state = distr_state_info ['state']

        if distr_state and distr_id:
            distr_state_info = self._normalize_dsi (distr_state_info)
            logging.debug ('Found dist id [%s] in state [%s]' % (distr_id, distr_state) )

        return distr_state_info


    def _get_distr_state_url (self, version, source_version = None, distr_type = None, client_filter = None):
        """
        forms an url to be used to get distr state info or request new distribution
        """
        logging.debug ('Reached _get_distr_state_url')
        logging.debug ('Request for url version [%s], type [%s], src_ver [%s]' % (version, distr_type, source_version) )

        if source_version:
            url = posixpath.join('dms-getver', 'rest', 'api', '1', 'distribution-difference')
            parms = {'initialFilters': client_filter, 'initialVersion': source_version, 'product': distr_type, 'targetFilters': client_filter, 'targetVersion': version}
        else:
            url = posixpath.join('dms-getver', 'rest', 'api', '1', 'distribution')
            parms = {'filter': client_filter, 'product': distr_type, 'version': version}
        logging.debug ('URL=[%s]' % url)
        return url, parms


    def _normalize_dsi (self, distr_state_info):
        """
        Normalize scattered output from different endpoints
        """
        logging.debug ('Reached _normalize_dsi')

        if 'initialFilters' in distr_state_info.keys ():
            logging.debug ('Diff distribution meta-info detected, add distOption=diff, filter, version')
            distr_state_info ['distOption'] = 'diff'
            distr_state_info ['filter'] = distr_state_info ['targetFilters']
            distr_state_info ['version'] = distr_state_info ['targetVersion']
        else:
            logging.debug ('Full distribution meta-info detected, only add distOption=full')
            distr_state_info ['distOption'] = 'full'
        return distr_state_info




#!/usr/bin/python2.7

import sys
#if sys.version[0]=='3':
#    from io import StringIO
#else:
#    import StringIO
import re
import json
import unittest
from datetime import datetime, timedelta
from collections import namedtuple
from oc_cdtapi.ForemanAPI import ForemanAPI, ForemanAPIError

class _Response(object):
    """
    Fake response object
    """

    def __init__(self, data):
        self.content = data
        self.text = json.dumps({"results": [{"id": 100, "name": "test_stand"}]}) 
        self.status_code = 200

    def json(self):
        return json.loads(self.content)


class _Session(object):
    """
    Fake requests session
    """

    def __init__(self, handler):
        self.handler = handler

    def get(self, req, params = None, **other):
        r = req
        if params is not None and len(params) > 0:
            r += '?' + '&'.join(map(lambda x: x [0]+'='+x [1], params.items()))
        return _Response(self.handler(r))

    def post(self, req, params = None, **other):
        r = req
        if params is not None and len(params) > 0:
            r += '?' + '&'.join(map(lambda x: x [0]+'='+x [1], params.items()))
        return _Response(self.handler(r))
    
    def delete(self, req, params = None, **other):
        r = req
        if params is not None and len(params) > 0:
            r += '?' + '&'.join(map(lambda x: x [0]+'='+x [1], params.items()))
        return _Response(self.handler(r))

    def put(self, req, params = None, **other):
        return _Response(self.handler(req))

class _ForemanAPI(ForemanAPI):

    def __init__(self, *args, **argv):
        self.web = _Session(self._read_url)
        self.root = "https://foreman"
        class_defaults = namedtuple("values", "exp_date location_id hostgroup deploy_on")
        exp_date = "01/01/2030"
        location_id = 5
        hostgroup = 11
        deploy_on = 1
        
        self.defs = class_defaults(exp_date, location_id, hostgroup, deploy_on)
        self.apiversion = 1

    def _read_url(self,url):
        """
        Overrides http request method
        """
        if re.match('.+\/hosts/test$',url):
            return '{"name": "test_stand", "uuid": "50391e80-afde-c4a5-c562-e5af02e5e449"}'
        elif re.match('.+\/hosts/test2$',url):
            return '{"name": "test_stand_2"}'
        elif re.match('.+\/puppetclasses/test_class', url):
            return '{"id": 101}'
        elif re.match('.+\/smart_class_parameters/1111', url):
            return '{"value": "default"}'
        elif re.match('.+\/hostgroups/1/puppetclasses', url):
            return '{"name": "puppet"}'
        elif re.match('.+\/subnets', url):
            return '{"total": 1, "subnetname": 64}'
        elif re.match('.+\/hosts/test/config_reports$', url):
            return '{"report_id": 102}'
        elif re.match('.+\/config_reports/101', url):
            return '{"result": "Success"}'
        elif re.match('.+\/usergroups.*%22.+%22$', url):
            return '{"results": [{"name":"NEW QA","id":2}]}'
        elif re.match('.+\/hostgroups$', url):
            return '{"results": [{"name":"docker-host","id":3}]}'
        elif re.match ('.+\/hosts\/test-host-name.*', url):
            return '{"result": "Success"}'
        elif re.match('.+\/organizations', url):
            return '{"total": 1, "results": [{"name": "CompanyName", "id": 1}]}'
        elif re.match('.+\/operatingsystems$', url):
            return '{"total": 1, "results": [{"description": "CentOS 7 Test", "id": 1}]}'
        elif re.match('.+\/operatingsystems/1/images$', url):
            return '{"total": 1, "results": [{"name": "CentOS 7.9", "uuid": "03366eb2-fc38-4813-970d-bd66c0b4cbf4"}]}'
        elif re.match('.+\/compute_resources/1/available_flavors$', url):
            return '{"total": 1, "results": [{"name": "cdt.1.4", "id": "13c5cccf-f907-4861-b367-bee86bec47cd"}]}'
        elif re.match('.+\/compute_resources/1$', url):
            return '{"description": "Test Provider", "compute_attributes": [{"attributes": {"tenant_id": "aeea0ca2b1ab449e86fd7b4295455ecf"}}]}'
        elif re.match('.+\/compute_resources/2$', url):
            return '{"description": "Test Provider", "compute_attributes": [{"attributes": {"availability_zone": "nova"}}]}'
        return '[]'

class TestForemanAPI(unittest.TestCase):

    def setUp(self):
        self.api = _ForemanAPI()
        self.json_object = json.dumps({"name": "custom_name_test", "is_owned_by": 100})
    
    def test_get_usergroup_id(self):
        group_id = self.api.get_usergroup_id("NEW QA")
        self.assertEqual(group_id, 2)

    def test_get_owner(self):
        user = self.api.get_owner("test")
        self.assertEqual(user, 100)

    def test_set_expiration(self):
        exp_date = self.api._set_expiration()
        self.assertEqual(exp_date, str((datetime.now() + timedelta(days=90)).strftime('%d/%m/%Y')))

    def test_create_host_no_hostname(self):
        with self.assertRaises(ForemanAPIError) as err:
            self.api.create_host()

    def test_create_host_default_values_no(self):
        with self.assertRaises(ForemanAPIError):
            self.api.create_host("test")

    def test_create_host_correct_values(self):
        try:
            self.api.create_host(custom_json=self.json_object)
        except ForemanAPIError:
            raise

    def test_get_host_info(self):
        info = self.api.get_host_info("test")
        self.assertEqual(info["name"], "test_stand")

    def test_get_host_info_unknown_host(self):
        info = self.api.get_host_info("test1")
        self.assertFalse(isinstance(info, dict))

    def test_puppet_class_info(self):
        info = self.api.puppet_class_info("test_class")
        self.assertEqual(info["id"], 101)

    def test_smart_class_info(self):
        info = self.api.smart_class_info(1111)
        self.assertEqual(info["value"], "default")

    def test_get_hostgroup_puppetclasses(self):
        classes = self.api.get_hostgroup_puppetclasses(1)
        self.assertEqual(classes["name"], "puppet")

    def test_get_subnets(self):
        subnets = self.api.get_subnets()
        self.assertEqual(subnets["subnetname"], 64)
    
    def test_get_host_reports(self):
        reports = self.api.get_host_reports("test")
        self.assertEqual(reports["report_id"], 102)
    
    def test_get_report(self):
        report = self.api.get_report(101)
        self.assertEqual(report["result"], "Success")
    
    def test_delete_host(self):
        try:
            self.api.delete_host("test")
        except ForemanAPIError:
            raise

    def test_hostgroup_found(self):
        hostgroup_id = self.api.get_hostgroup_id("docker-host")
        self.assertEqual(hostgroup_id, 3)
    
    def test_hostgroup_not_found(self):
        hostgroup_id = self.api.get_hostgroup_id("docker")
        self.assertEqual(hostgroup_id, None)

    def test_get_organization_id(self):
        organization_id = self.api.get_organization_id("CompanyName")
        self.assertEqual(organization_id, 1)

    def test_get_organization_id_not_found(self):
        organization_id = self.api.get_organization_id("CompanyName2")
        self.assertIsNone(organization_id)
    
    def test_set_host_expiry(self):
        self.api.set_host_expiry('test-host-name','2222-01-22')

    def test_get_image_uuid(self):
        uuid = self.api.get_image_uuid("CentOS 7 Test", "CentOS 7.9")
        self.assertEqual("03366eb2-fc38-4813-970d-bd66c0b4cbf4", uuid)
    
    def test_get_image_uuid_not_found(self):
        uuid = self.api.get_image_uuid("CentOS 7", "CentOS 7")
        self.assertIsNone(uuid)
    
    def test_get_flavor_id(self):
        flavor_id = self.api.get_flavor_id(1, "cdt.1.4")
        self.assertEqual("13c5cccf-f907-4861-b367-bee86bec47cd", flavor_id)
    
    def test_get_flavor_id_not_found(self):
        flavor_id = self.api.get_flavor_id(1, "cdt.1.5")
        self.assertIsNone(flavor_id)
    
    def test_get_tenant_id(self):
        tenant_id = self.api.get_tenant_id(1)
        self.assertEqual("aeea0ca2b1ab449e86fd7b4295455ecf", tenant_id)
    
    def test_get_tenant_id_not_found(self):
        tenant_id = self.api.get_tenant_id(2)
        self.assertIsNone(tenant_id)
    
    def test_get_host_uuid(self):
        uuid = self.api.get_host_uuid("test")
        self.assertEqual("50391e80-afde-c4a5-c562-e5af02e5e449", uuid)
    
    def test_get_host_uuid(self):
        uuid = self.api.get_host_uuid("test2")
        self.assertIsNone(uuid)




# requires python-requests rpm package
import doctest
import os
import shutil  # this required to copy data between file objects
import posixpath

import requests

import sys

if sys.version_info.major == 2:
    strtype = basestring
elif sys.version_info.major == 3:
    strtype = str

class HttpAPIError(Exception):

    def __init__(self, code=0, url='', resp=None, text=''):
        self.code = code
        self.url = url
        self.resp = resp
        self.text = text


    def __str__(self):
        return self.text + ': Code ' + str(self.code) + ' ' + self.url


class HttpAPI(object):
    """ Base class for implementing HTTP API """
    # This attributes may be re-defined in your child classes
    _error = HttpAPIError  # Error to throw if something goes wrong
    raise_exception_low = 200  # Rise exception if return code is < 200
    raise_exception_high = 399  # Raise exception if return code is > 399
    _env_prefix = None  # If set, we read prefix_USER, prefix_PASSWORD and prefix_URL
    # if root, user and auth was not provided to class constructor
    _env_url = '_URL'
    _env_user = '_USER'
    _env_auth = '_PASSWORD'

    def __init__(self, root=None, user=None, auth=None, readonly=False, anonymous=False):
        """
        :param root: Root URL (uses *_URL by default)
        :param user: Username (uses *_USER by default)
        :param auth: Password (uses *_PASSWORD by default)
        :param readonly: sets readonly property, should be explictly supported by child 
        :param anonymous: ignore username/password provided and use anonymous requests

        >>> import os
        >>> from oc_cdtapi.API import HttpAPI
        >>> def test(port):
        ...     HttpAPI._env_prefix='HTTP'
        ...     os.environ['HTTP_URL'] = 'http://127.0.0.1:' + str(port)
        ...     api = HttpAPI()
        ...     api.get('testinit')
        ...
        >>> from oc_cdtapi.TestServer import test_wrapper
        >>> test_wrapper(test)
        "GET /testinit HTTP/1.1" 200 -
        >>> os.environ[ 'HTTP_URL' ] = "https://127.0.0.1/";
        >>> HttpAPI._env_prefix = 'HTTP';
        >>> obj_api = HttpAPI();
        >>> obj_api.root
        'https://127.0.0.1/'
        >>> obj_api.web.verify
        False
        """

        self.readonly = readonly
        if self._env_prefix != None:
            if not root: root = os.getenv(self._env_prefix + self._env_url)
            if not user: user = os.getenv(self._env_prefix + self._env_user)
            if not auth: auth = os.getenv(self._env_prefix + self._env_auth)
        if not root: raise ValueError('Server URL for service [%s] not set' % self._env_prefix)
        self.root = root
        self.web = requests.Session()
        if user != None and not anonymous: self.web.auth = (user, auth)

        if self.root.startswith( "https:" ):
            self.web.verify = False;


    def set_readonly(self, readonly=True):
        self.readonly = readonly


    def re(self, req):
        """ Constructs request URL from request name """
        return posixpath.join(self.root, req)


    def pp(self, resp, write_to=None, stream=False, **kvarg):
        """ Post-processes response

        New feature! If write_to parameter is given to any method, that returns response
        object, request result will be written to fd or filename specified in write_to.

        If stream == True, file will be written on the fly, without reading all the data

        >>> import sys
        >>> import io
        >>> class FakeResp(object):
        ...     raw = io.BytesIO(b'Read data')
        ...     content = b'Read data!'
        ...     status_code = 200
        ...
        >>> api = HttpAPI('http://127.0.0.1')
        >>> fake_resp = FakeResp()
        >>> read_fd = io.BytesIO()
        >>> r = api.pp(fake_resp, stream = True, write_to = read_fd)
        >>> r == fake_resp
        True
        >>> if sys.version_info.major == 2:
        ...     value = read_fd.getvalue()
        ... else:
        ...     value = read_fd.getvalue().decode()
        >>> value
        'Read data'
        >>> read_fd = io.BytesIO()
        >>> r = api.pp(fake_resp, stream = False, write_to = read_fd)
        >>> r == fake_resp
        True
        >>> if sys.version_info.major == 2:
        ...     value = read_fd.getvalue()
        ... else:
        ...     value = read_fd.getvalue().decode()
        >>> value
        'Read data!'
        >>> fake_resp.raw = io.BytesIO(b'Read data2')
        >>> import tempfile
        >>> tmpfile = tempfile.mktemp()
        >>> if os.path.exists(tmpfile): os.unlink(tmpfile)
        >>> r = api.pp(fake_resp, stream = True, write_to = tmpfile)
        >>> r == fake_resp
        True
        >>> open(tmpfile, 'r').read()
        'Read data2'
        >>> os.unlink(tmpfile)
        >>> r = api.pp(fake_resp, stream = False, write_to = tmpfile)
        >>> r == fake_resp
        True
        >>> open(tmpfile, 'r').read()
        'Read data!'
        >>> os.unlink(tmpfile)

        """

        if resp.status_code < self.raise_exception_low or resp.status_code > self.raise_exception_high:
            raise self._error(resp.status_code, resp.url, resp, 'Error making request to server')
        if write_to != None:
            if isinstance (write_to, strtype):
                fd = open(write_to, 'wb')
            else:
                fd = write_to
            if not stream:
                fd.write(resp.content)
            else:
                shutil.copyfileobj(resp.raw, fd)
            fd.flush()
            if isinstance (write_to, strtype): fd.close()
            del fd
        return resp


    def __kvarg(self, kvarg):
        kvarg = kvarg.copy()
        if 'write_to' in kvarg: del kvarg['write_to']
        return kvarg


    def get(self, req, params=None, files=None, data=None, headers=None, **kvarg):
        """Sends GET request

        >>> def test(port):
        ...     api = HttpAPI('http://127.0.0.1:' + str(port))
        ...     api.get('getrequest')
        ...
        >>> from oc_cdtapi.TestServer import test_wrapper
        >>> test_wrapper(test)
        "GET /getrequest HTTP/1.1" 200 -

        >>> import io
        >>> def test(port):
        ...     api = HttpAPI('http://127.0.0.1:' + str(port))
        ...     bio = io.BytesIO()
        ...     api.get('getrequest', write_to = bio)
        ...     if bio.getvalue() == b'hello, world!': exit(0)
        ...     exit(123)
        ...
        >>> test_wrapper(test, data = 'hello, world!', ret = True)
        "GET /getrequest HTTP/1.1" 200 -
        0

        """

        resp = self.web.get(self.re(req), params=params, data=data, files=files, headers=headers, **self.__kvarg(kvarg))
        return self.pp(resp, **kvarg)


    def post(self, req, params=None, files=None, data=None, headers=None, **kvarg):
        """Sends POST request"""
        resp = self.web.post(self.re(req), params=params, data=data, files=files, headers=headers, **kvarg)
        return self.pp(resp, **kvarg)


    def put(self, req, params=None, files=None, data=None, headers=None, **kvarg):
        """Sends PUT request
        >>> def test(port):
        ...     api = HttpAPI('http://127.0.0.1:' + str(port))
        ...     api.put('getrequest')
        ...
        >>> from oc_cdtapi.TestServer import test_wrapper
        >>> test_wrapper(test)
        "PUT /getrequest HTTP/1.1" 200 -

        """

        resp = self.web.put(self.re(req), params=params, data=data, files=files, headers=headers, **kvarg)
        return self.pp(resp, **kvarg)


    def delete(self, req, params=None, files=None, data=None, headers=None, **kvarg):
        """Sends DELETE request"""
        resp = self.web.delete(self.re(req), params=params, data=data, files=files, headers=headers, **kvarg)
        return self.pp(resp, **kvarg)

    def head(self, req, params=None, files=None, data=None, headers=None, **kvarg):
        """Sends HEAD request"""
        resp = self.web.head(self.re(req), params=params, data=data, files=files, headers=headers, **kvarg)
        return self.pp(resp, **kvarg)


# Two shortcuts to deal with XML without having full XML support
def get_xml_tag(config, tag):
    """Gets tag value from xml config"""
    return ((config.split('<' + tag + '>')[1]).split('</' + tag + '>')[0])


def edit_xml_tag(config, tag, value):
    """Sets tag value in xml config. Works fine if there is only single such tag in XML"""
    config = config.split('<' + tag + '>')[0] + '<' + tag + '>' + value + '</' + tag + '>' + \
             config.split('</' + tag + '>')[1]
    return config


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# This module contains little test HTTP server for checking HTTP API based modules

import sys
from multiprocessing import Process
from requests import Session
import doctest

if sys.version_info.major == 3:
    from http.server import HTTPServer, BaseHTTPRequestHandler
else:
    from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler

class HTTPRequestHandler(BaseHTTPRequestHandler):

    def process(self,tpy):
        self.send_response(self.server.ret_code)
        if self.server.data_headers:
            for header in self.server.data_headers:
                self.send_header(header)
        self.end_headers()
        if 'data_to_send' in self.server.__dict__ and self.server.data_to_send:
            dts = self.server.data_to_send
            if sys.version_info.major == 3:
                dts = bytes (dts, 'UTF-8')
            self.wfile.write (dts)


    def do_PUT(self):
        self.process('PUT')


    def do_GET(self):
        self.process('GET')


    def do_POST(self):
        self.process('POST')


    def do_DELETE(self):
        self.process('DELETE')

    def do_HEAD(self):
        self.process('HEAD')


    def log_message(self,fmt, *args):
        print((fmt % args))



def test_server(port, rc = 200, data = None, headers = []):
    srv = HTTPServer(('127.0.0.1',port),HTTPRequestHandler)
    srv.ret_code = rc
    srv.data_to_send = data
    srv.data_headers = headers
    return srv


def test_function(port, request='', code = 0, data_to_expect = None):
    resp = Session().get('http://127.0.0.1:' + str(port) + request)
    if data_to_expect:
        received = resp.content.decode('UTF-8')
        if received != data_to_expect: exit(254)
    exit(code)


def test_wrapper(func = test_function, num_requests = 1, rc=200, ret=False, data=None, headers=None, *args, **kvargs):
    """
    >>> from TestServer import test_wrapper
    >>> test_wrapper()
    "GET / HTTP/1.1" 200 -
    >>> test_wrapper(rc=307)
    "GET / HTTP/1.1" 307 -
    >>> test_wrapper(rc=404, ret = True, code=112)
    "GET / HTTP/1.1" 404 -
    112
    >>> test_wrapper(ret = True, data_to_expect = 'hello', data = 'hello')
    "GET / HTTP/1.1" 200 -
    0
    >>> test_wrapper(ret = True, data_to_expect = 'hello', data = 'hi')
    "GET / HTTP/1.1" 200 -
    254
    """

    if not 'port' in kvargs: kvargs['port'] = 65111
    srv = test_server(kvargs['port'], rc, data, headers)
    proc = Process(group=None, target=func, name=None, args=args, kwargs=kvargs)
    proc.daemon = 1
    proc.start()
    for req in range(0, num_requests): srv.handle_request()
    proc.join(5)
    del srv
    if ret: return proc.exitcode


if __name__ == "__main__":
    import doctest
    doctest.testmod()






import logging, unittest
from oc_cdtapi.DmsGetverAPI import DmsGetverAPI



class _DmsGetverAPI (DmsGetverAPI):


    def __init__ (self):
        super (_DmsGetverAPI, self).__init__ ('fake url', 'fake user', 'fake password')
        self.req_counter = 0


    def _fake_state_info (self, distr_id = None, distr_state = None, distr_filename = None):
        distr_state_info = {}
        distr_state_info ['id'] = distr_id
        distr_state_info ['state'] = distr_state
        distr_state_info ['fileName'] = distr_filename
        return distr_state_info


    def _fake_state_info_diff (self, distr_id = None, distr_state = None, distr_filename = None):
        distr_state_info = {}
        distr_state_info ['id'] = distr_id
        distr_state_info ['state'] = distr_state
        distr_state_info ['fileName'] = distr_filename
        distr_state_info ['initialFilters'] = 'FILTER'
        distr_state_info ['targetFilters'] = 'FILTER'
        distr_state_info ['initialVersion'] = 'VERSION'
        distr_state_info ['targetVersion'] = 'VERSION'
        return distr_state_info


    def _fake_gav (self):
        fake_gav = {}
        fake_gav ['groupId'] = 'groupId'
        fake_gav ['artifactId'] = 'artifactId'
        fake_gav ['version'] = 'version'
        fake_gav ['packaging'] = 'packaging'
        return fake_gav


    def get (self, url, params=None):
        resp = FakeResp ()
        if (url == 'dms-getver/rest/api/1/distribution' and params and params.get('version') == '09.99.99.99') or \
           url == 'dms-getver/rest/api/1/distribution/id:99' or \
           url == 'dms-getver/rest/api/1/distribution-difference/id:99':
            if self.req_counter == 0:
                resp.status_code = 200
                resp.content = None
                resp.jdata = self._fake_state_info (99, 'PROCESSING', None)
                self.req_counter += 1
            else:
                resp.status_code = 200
                resp.content = None
                resp.jdata = self._fake_state_info (99, 'READY', None)
                self.req_counter += 1
        elif url == 'dms-getver/rest/api/1/distribution-difference':
            resp.status_code = 200
            resp.content = None
            resp.jdata = self._fake_state_info_diff (99, 'PROCESSING', None)
        elif (url == 'dms-getver/rest/api/1/distribution' and params and params.get('version') == '07.77.77.77'):
            resp.status_code = 404
            jdata = {}
            jdata ['code'] = 'DMS-GETVER-40001'
            resp.jdata = jdata
        elif url == 'dms-getver/distribution/id:99/download':
            if self.req_counter < 3:
                resp.status_code = 200
                resp.content = 'ABCDEF0123456789'
                resp.jdata = self._fake_state_info (99, 'READY', None)
            else:
                resp.status_code = 404
                resp.content = None
                resp.jdata = None
        elif url == 'dms-getver/distribution/id:88/download':
            resp.status_code = 200
            resp.jdata = self._fake_state_info (99, 'PROCESSING', None)
        elif url == 'dms-getver/distribution/id:77/download':
            resp.status_code = 404
            resp.jdata = None
        elif url == 'dms-getver/rest/api/1/distribution/id:99/gav':
            resp.status_code = 200
            resp.jdata = self._fake_gav ()
        elif url.find ('/log') != -1:
            if url.find ('99') != -1:
                resp.status_code = 200
                resp.text = 'dms log data'
            else:
                resp.status_code = 404
                resp.text = 'log not found'
        return resp


    def post (self, url, json = None):
        resp = FakeResp ()
        if url == 'dms-getver/rest/api/1/distribution' or \
           url == 'dms-getver/rest/api/1/distribution-difference':
            if 'version' in json.keys ():
                version = json ['version']
            elif 'targetVersion' in json.keys ():
                version = json ['targetVersion']
            if version == '09.99.99.99':
                resp.status_code = 200
                resp.content = None
                resp.jdata = self._fake_state_info (99, 'INITIATED', None)
            elif version == '08.88.88.88':
                resp.status_code = 500
                resp.content = None
        return resp




class FakeResp (object):


    def __init__ (self):
        self.jdata = None


    def status_code (self):
        return None


    def content (self):
        return None


    def json (self):
        if self.jdata:
            return self.jdata
        else:
            raise ValueError


    def text (self):
        return None




class TestDmsGetverAPI (unittest.TestCase):


    def test_none (self):
        self.assertTrue (True)


    def test_dms_log_ok (self):
        da = _DmsGetverAPI ()
        distr_option = 'full'
        log = da.get_dms_log (99, distr_option)
        self.assertEqual (log, 'dms log data')


    def test_dms_log_404 (self):
        da = _DmsGetverAPI ()
        distr_option = 'full'
        log = da.get_dms_log (77, distr_option)
        self.assertIsNone (log)


    def test_create_distr_request_full (self):
        da = _DmsGetverAPI ()
        version = '09.99.99.99'
        distr_type = 'CARDS'
        client_filter = 'FILTER'
        distr_state_info = da.create_distr_request (version = version, distr_type = distr_type, client_filter = client_filter)
        state = distr_state_info ['state']
        self.assertEqual (state, 'INITIATED')


    def test_create_distr_request_diff (self):
        da = _DmsGetverAPI ()
        source_version = '08.88.88.88'
        version = '09.99.99.99'
        distr_type = 'CARDS'
        client_filter = 'FILTER'
        distr_state_info = da.create_distr_request (version = version, source_version = source_version, distr_type = distr_type, client_filter = client_filter)
        state = distr_state_info ['state']
        self.assertEqual (state, 'INITIATED')


    def test_create_distr_request_filterless (self):
        da = _DmsGetverAPI ()
        source_version = '08.88.88.88'
        version = '09.99.99.99'
        distr_type = 'CARDS'
        client_filter = 'FILTER'
        distr_state_info = da.create_distr_request (version = version, source_version = source_version, distr_type = distr_type, client_filter = None)
        state = distr_state_info ['state']
        self.assertEqual (state, 'INITIATED')


    def test_create_distr_request_500 (self):
        da = _DmsGetverAPI ()
        version = '08.88.88.88'
        distr_type = 'CARDS'
        client_filter = 'FILTER'
        distr_state_info = da.create_distr_request (version = version, distr_type = distr_type, client_filter = None)
        state = distr_state_info ['state']
        self.assertEqual (state, 'HTTP/500')


    def test_dumb_404_dumb0 (self):
        da = _DmsGetverAPI ()
        jdata = {}
        jdata ['code'] = 'DMS-GETVER-40001'
        resp = FakeResp ()
        resp.status_code = 404
        resp.jdata = jdata
        is_dumb = da._dumb_404 (resp)
        self.assertTrue (is_dumb)


    def test_dumb_404_dumb1 (self):
        da = _DmsGetverAPI ()
        jdata = {}
        jdata ['code'] = 'blahblahblah'
        resp = FakeResp ()
        resp.status_code = 404
        resp.jdata = jdata
        is_dumb = da._dumb_404 (resp)
        self.assertFalse (is_dumb)


    def test_dumb_404_real (self):
        da = _DmsGetverAPI ()
        jdata = {}
        resp = FakeResp ()
        resp.status_code = 404
        is_dumb = da._dumb_404 (resp)
        self.assertFalse (is_dumb)


    def test_get_distr_full_ok (self):
        da = _DmsGetverAPI ()
        da.req_counter = 1
        distr_id = 99
        distr_option = 'full'
        distr, distr_state_info = da.get_distr (distr_id, distr_option)
        self.assertEqual (distr, 'ABCDEF0123456789')


    def test_get_distr_full_notready (self):
        da = _DmsGetverAPI ()
        da.req_counter = 1
        distr_id = 88
        distr_option = 'full'
        distr, distr_state_info = da.get_distr (distr_id, distr_option)
        self.assertIsNone (distr)


    def test_get_distr_full_404 (self):
        da = _DmsGetverAPI ()
        da.req_counter = 4
        distr_id = 99
        distr_option = 'full'
        distr, distr_state_info = da.get_distr (distr_id, distr_option)
        self.assertIsNone (distr)


    def test_get_distr_diff_ok (self):
        da = _DmsGetverAPI ()
        da.req_counter = 1
        distr_id = 99
        distr_option = 'diff'
        distr, distr_state_info = da.get_distr (distr_id, distr_option)
        self.assertEqual (distr, 'ABCDEF0123456789')


    def test_get_distr_state_info_full (self):
        da = _DmsGetverAPI ()
        version = '09.99.99.99'
        distr_type = 'CARDS'
        client_filter = 'FILTER'
        distr_state_info = da.get_distr_state_info (version = version, distr_type = distr_type, client_filter = client_filter)
        state = distr_state_info ['state']
        self.assertEqual (state, 'PROCESSING')


    def test_get_distr_state_info_full_dumb404 (self):
        da = _DmsGetverAPI ()
        version = '07.77.77.77'
        distr_type = 'CARDS'
        client_filter = 'FILTER'
        distr_state_info = da.get_distr_state_info (version = version, distr_type = distr_type, client_filter = client_filter)
        state = distr_state_info ['state']
        self.assertEqual (state, 'NOTFOUND')


    def test_get_distr_state_info_diff (self):
        da = _DmsGetverAPI ()
        version = '09.99.99.99'
        source_version = '08.88.88.88'
        distr_type = 'CARDS'
        client_filter = 'FILTER'
        distr_state_info = da.get_distr_state_info (version = version, source_version = source_version, distr_type = distr_type, client_filter = client_filter)
        state = distr_state_info ['state']
        self.assertEqual (state, 'PROCESSING')


    def test_get_distr_state_info_byid_full (self):
        da = _DmsGetverAPI ()
        da.req_counter = 1
        distr_id = 99
        distr_option = 'full'
        distr_state_info = da.get_distr_state_info_byid (distr_id, distr_option)
        state = distr_state_info ['state']
        self.assertEqual (state, 'READY')


    def test_get_distr_state_info_byid_diff (self):
        da = _DmsGetverAPI ()
        da.req_counter = 1
        distr_id = 99
        distr_option = 'diff'
        distr_state_info = da.get_distr_state_info_byid (distr_id, distr_option)
        state = distr_state_info ['state']
        self.assertEqual (state, 'READY')


    def test_get_dms_gav (self):
        da = _DmsGetverAPI ()
        gav = da.get_dms_gav (99, 'full')
        self.assertEqual (gav, 'groupId:artifactId:version:packaging')


    def test_wait_for_state_ok (self):
        da = _DmsGetverAPI ()
        da.wait_state_sleep = 1
        distr_id = 99
        distr_option = 'full'
        distr_state_info = da.wait_for_state (distr_id, distr_option)
        state = distr_state_info ['state']
        self.assertEqual (state, 'READY')


    def test_wait_for_state_timeout (self):
        da = _DmsGetverAPI ()
        da.wait_state_sleep = 5
        da.wait_state_timeout = -1
        distr_id = 99
        distr_option = 'full'
        distr_state_info = da.wait_for_state (distr_id, distr_option)
        state = distr_state_info ['state']
        self.assertEqual (state, 'TIMEOUT')


if __name__ == '__main__':
    unittest.main()




import doctest
import os
import re
from xml.etree import ElementTree
import posixpath
import logging

import requests
from .API import HttpAPI, HttpAPIError

import sys
if sys.version_info.major == 2:
    strtype = basestring
elif sys.version_info.major == 3:
    strtype = str

# Nexus url should include "/nexus" ending
# GAV has the following format: groupid:artifactid:version[:packaging[:clasifier]]
# This interface should be considered low-level

class NexusAPIError(HttpAPIError):
    def __str__(self):
        return 'Nexus error: ' + self.text + ': Code ' + str(self.code) + ' ' + self.url


def parse_gav(gav, relaxed=False):
    """
    Parses GAV string and returns dict with its components suitable for nexus calls
    """

    if isinstance (gav, dict):
        if any([relaxed, set(['a', 'g', 'v']).issubset(set(gav.keys()))]): return gav
        raise ValueError("Expected string or dict with g/a/v but got dict")

    if relaxed and (gav == None or gav == ''): return {}
    if not isinstance (gav, strtype): raise ValueError("Expected string but got " + str(type(gav)))
    gav_list = gav.split(':')
    if len(gav_list) < 3 and not relaxed: raise ValueError("GroupId, ArtifactId and Version are mandatory")
    if len(gav_list) > 5: raise ValueError("Incorrect GAV format")
    gav_dict = {}
    for i, v in enumerate(gav_list): gav_dict[['g', 'a', 'v', 'p', 'c'][i]] = v
    return gav_dict


def gav_to_path(gav, relaxed=False):
    """
    Converts GAV to repository path. Gav can be either string or dict from parse_gav function
    """

    if isinstance (gav, strtype):
        gav = parse_gav(gav, relaxed)

    if not relaxed and not set(['g', 'a', 'v']).issubset(set(gav.keys())):
        raise ValueError('ArtifactId, GroupId and Version are mandatory!')

    url = ''
    group_path = gav['g'].replace('.', posixpath.sep) if "g" in gav else ""
    artifact_path = gav_to_artifact_path(gav) if "a" in gav else ""
    url = posixpath.sep.join(filter(None, [group_path, artifact_path]))
    return url


def gav_to_artifact_path(gav):
    """ Converts gav to name like artifactid/version/filename """
    if isinstance (gav, strtype):
        gav = parse_gav(gav, relaxed=True)

    path = gav['a']

    if 'v' in gav:
        path = posixpath.join(path, gav['v'], gav_to_filename(gav))

    return path


def gav_to_filename(gav):
    """ Converts gav to name like artifactid-version.packaging """
    if isinstance (gav, strtype):
        gav = parse_gav(gav, relaxed=True)

    if not set(['a', 'v']).issubset(set(gav.keys())):
        raise ValueError("Artifactid and version required to make path")

    filename = '-'.join(
            filter(None, map(lambda x: gav.get(x), ['a', 'v', 'c'])))

    return '.'.join([filename, gav.get('p', 'jar')])

def gav_to_str(parsed_gav):
    """ Converts parsed GAV back to colon-separated string """
    required_keys=["g", "a", "v"]

    if not set(required_keys).issubset(set(parsed_gav.keys())):
        raise ValueError("At least groupid, artifactid and version should be presented in parsed gav")

    _result = ':'.join(map(lambda x: parsed_gav.get(x), required_keys))

    if not 'p' in parsed_gav.keys():
        return _result

    return ':'.join([_result] + list(filter(None, map(lambda x: parsed_gav.get(x), ['p', 'c']))))


def pom_from_gav(gav, no_packaging=True):
    """ 
    Generates pom file from gav
    """

    #  1. perhaps this is a good place to make it with:
    #       - pkg_resource - to get template itself 
    #       - jinja2.Template() - to fill it with values
    #     but it was not implemented to get rid of extra libraries dependency
    #  2. constructing it with xml.etree.ElementTree looks more correct
    #     but code will be longer than a template. 
    #       - ElementTree has many issues with constructing main <project> tag
    #       - Earlier versions of ElementTree are unable to write <?xml?> declaration to result output
    #       - Python2 and Python3 versions of ElementTree have different behaviour with return types of 'tostring'
    #           - For Python3 it is necessary to specify encoding='unicode', 
    #             otherwise result type will be 'bytes', not 'strtype'
    #           - Python2 version of ElementTree does not support 'unicode' encoding. 
    pom_template = '''<?xml version="1.0" encoding="UTF-8"?>
      <project xsi:schemaLocation="http://maven.apache.org/POM/4.0.0 http://maven.apache.org/xsd/maven-4.0.0.xsd" 
            xmlns="http://maven.apache.org/POM/4.0.0" 
            xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
        <modelVersion>4.0.0</modelVersion>
        <groupId>{g}</groupId>
        <artifactId>{a}</artifactId>
        <version>{v}</version>
        {pkg}
      </project>
    '''

    if isinstance (gav, strtype):
        gav = parse_gav(gav)
    else:
        gav = gav.copy()

    if no_packaging or 'p' not in gav.keys():
        gav['pkg'] = ''
    else:
        gav['pkg'] = '<packaging>' + gav['p'] + '</packaging>'

    return pom_template.format(**gav)

def gav_from_path(path, dictionary=True):
    """
    Generate GAV from path to artifact.
    Path should be relative to main entrypoint
    And not should not contain repositoryId. Example:
    """
    generated_gav = {}
    path = path.strip('.').strip(posixpath.sep)
    path_parts = path.split(posixpath.sep)
    generated_gav['g'] = '.'.join(path_parts[:-3])
    generated_gav['a'] = path_parts[-3]
    generated_gav['v'] = path_parts[-2]

    file_name = posixpath.basename(path)
    _av_part = '{a}-{v}'.format(**generated_gav)
    _cp_part = None

    # Actually if filename does not start with '_av_part' in MVN - 
    # it is surely a bug, but we shall try to act gracefully
    if not file_name.startswith(_av_part):
        return generated_gav if dictionary else gav_to_str(generated_gav)

    _cp_part = file_name.replace(_av_part, '', 1).strip('-').strip('.')

    if not _cp_part:
        # nothing to parse for packaging and classifier
        return generated_gav if dictionary else gav_to_str(generated_gav)
        
    classifier, file_ext = posixpath.splitext(_cp_part)

    if file_ext:
        # So we have packaging info and classifier
        generated_gav['p'] = file_ext.strip('.')
        generated_gav['c'] = classifier
    else:
        # otherwise classifier is absent, so there is packaging there
        generated_gav['p'] = classifier

    return generated_gav if dictionary else gav_to_str(generated_gav)

class NexusAPI(HttpAPI):
    _error = NexusAPIError
    _env_prefix = 'MVN'
    _env_upload_repo = '_UPLOAD_REPO'
    _env_download_repo = '_DOWNLOAD_REPO'

    codepage = 'utf-8'
    codepage_errors = 'replace'

    def __init__(self, root=None, user=None, auth=None,
                 readonly=False, anonymous=False, upload_repo=None, download_repo=None):
        """
        :param root: service URL
        :type root: str
        :param user: username
        :type user: str
        :param auth: password or authentication token
        :type auth: str
        :param readonly: do not try to put anything to the server
        :type readonly: bool
        :param anonymous: forse anonymous connection, even if _USER and _PASSWORD are specified in
        :param anonymous: the environment
        :type anonymous: bool
        :param upload_repo: repository to upload artifacts to
        :type repo: str
        :param download_repo: repository to download artifacts from
        :type download_repo: str
        """
        super(NexusAPI, self).__init__(root, user, auth, readonly, anonymous)
        self.__upload_repo = upload_repo
        self.__download_repo = download_repo
    
    @property
    def repo_default(self):
        """
        Return default repo for different systems (former "public")
        """
        if self.is_artifactory: 
            return "maven-virtual"

        return "public"

    @property
    def is_artifactory(self):
        """
        Test if we are working with Artifactory, not with Nexus
        """
        return self.root.rstrip(posixpath.sep).endswith(posixpath.sep + "artifactory")

    @property
    def is_nexus(self):
        """
        Test if we are working with Nexus, not with Artifactory
        """
        return self.root.rstrip(posixpath.sep).endswith(posixpath.sep + "nexus")

    def gav_get_url(self, gav, repo=None):
        """
        Get full artifact URL.
        :param gav: gav
        :type gav: str, dict
        :param repo: repo
        :type repo: str
        """
        # here is nice place to use 'urllib.urljoin'
        # but implemented ugly via 'posixpath' to get rid of extra dependencies
        if not gav:
            raise ValueError("GAV is mandatory")

        return posixpath.join(self.root, self.__gav_get_url(gav, repo=repo))

    def __gav_get_url(self, gav, repo, relaxed=False):
        """
        Full sub-path for get/post requests to 'gav'
        :param gav: gav
        :type gav: str, dict
        :param repo: repo
        :type repo: str
        """
        _rq = posixpath.join(
                repo if repo else self.repo_default, 
                gav_to_path(gav, relaxed=relaxed) if gav else "")

        if self.is_nexus: 
            _rq = posixpath.join("content", "repositories", _rq)

        return _rq.rstrip(posixpath.sep)

    def cat(self, gav, repo=None, binary=False, response=False, stream=False, encoding=None, enc_errors=None,
            write_to=None, rest_call=False, **argv):
        """
        Gets data from maven repo
        :param gav: GAV
        :type gav: str
        :param repo: repository to search artifact in
        :type repo: str
        :param binary: return binary artifact data, without decoding to text
        :type binary: bool
        :param response: return HTTP response itself instead of artifact data
        :type response: bool
        :param stream: open file stream to further process artifact data, useful for large files
        :type stream: boolean
        :param enc_errors: behaviour on text decoding/encoding errors while processing artifact data
        :type enc_errors: str
        :param write_to: path or file_object to write artifact data to instead of returning it as result
        :type write_to: str, file_obj, ...
        :param rest_call: try to search artifact via REST API instead of getting directly from repository
        :param rest_call: WARNING: this flag is ignored for Artifactory
        :type rest_call: bool
        """

        # we set binary to true to ignore codepage issues
        params = parse_gav(gav)
        if not repo: repo = self.__download_repo
        if not repo: repo = os.getenv(self._env_prefix + self._env_download_repo, self.repo_default)
        params['r'] = repo

        # workaround Nexus REST API can't work with plain text files, so we use direct download instead
        if rest_call and self.is_nexus:
            r = self.get(posixpath.join('service', 'local', 'artifact', 'maven', 'content'), params, stream=stream, write_to=write_to, **argv)
        else:
            _rq = self.__gav_get_url(gav, repo=params['r'])
            r = self.get(_rq, stream=stream, write_to=write_to, **argv)

        if response or write_to: return r
        if binary: return r.content

        if not encoding: encoding = self.codepage
        if not enc_errors: enc_errors = self.codepage_errors
        if not enc_errors: enc_errors = 'strict'
        if encoding: return r.content.decode(encoding, enc_errors)

        return r.text

    def exists(self, gav, repo=None, rest_call=False):
        """
        Checks if artifact exists in repo
        :param gav: gav
        :type gav: str
        :param repo: repository to search artifact in:
        :type repo: str
        :param rest_call: search with REST API instead of direct path
        :param rest_call: WARNING: this flag is for Nexus only
        :type rest_call: bool
        """
        params = parse_gav(gav)

        if not repo: repo = self.__download_repo
        if not repo: repo = os.getenv(self._env_prefix + self._env_download_repo, self.repo_default)
        params['r'] = repo

        if rest_call and self.is_nexus:
            try:
                r = self.get(posixpath.join('service', 'local', 'artifact', 'maven', 'redirect'), params, allow_redirects=False)
            except NexusAPIError as e:
                if e.code == 404: return False
                raise (e)
            if r.status_code == 307: return True
            raise self._error(r.status_code, r.url, r, 'Incorrect response code - only 404 and 307 are expected')
        
        _req = self.__gav_get_url(gav, repo=params['r'])
        try:
            r = self.head(_req)
        except NexusAPIError as e:
            if e.code == 404: return False
            raise (e)
        if r.status_code == 200: return True

        raise self._error(r.status_code, r.url, r, 'Incorrect response code - only 404 and 200 are expected')

    def upload(self, gav, repo=None, data=None, pom=None, metadata=False):
        """ Puts data into maven repo under gav
        pom can be set to actual pom content or to True for automatic pom generation
        :param gav: GAV
        :type gav: str
        :param repo: repository to upload to
        :type repo: str
        :param data: data ot upload
        :type data: str, bytes, file-like object, ...
        :param pom: upload POM. POM data may be provided as value. If 'True' provided - upload auto-generated POM
        :type pom: str, bool
        :param metadata: update metadata
        :param metadata: WARNING: this flag is valid for Nexus only and ignored for others
        :type metadata: bool
        """
        if not repo: repo = self.__upload_repo
        if not repo: repo = os.getenv(self._env_prefix + self._env_upload_repo)
        if not repo: raise (ValueError("You must provide repo name"))

        # Raise exception manually instead of
        # making data mandatory for backwards-compatibility
        if data == None: raise (ValueError("No data to upload"))

        # Requests uses 'multipart/form-data' header by default, which is not supported by Artifactory.
        custom_headers = {'Content-Type': 'application/binary'}

        # TODO: temporal hack, to be removed
        #       'repo' may be "repositories/id/sub" or longer. This case we need second component only.
        if repo.startswith('repositories' + posixpath.sep): repo = repo.split(posixpath.sep)[1]

        url = self.__gav_get_url(gav, repo=repo)
        params = parse_gav(gav)
        params['p'] = 'pom'
        pom_url = self.__gav_get_url(params, repo=repo)

        # we need 'True' exactly
        if pom is True: 
            # don't try to upload auto-generated pom if there is existing one
            if self.exists(params, repo=repo): pom = False
            else: pom = pom_from_gav(gav, no_packaging=True)

        # uploading generated or provided pom
        if pom: self.put(pom_url, data=pom, headers=custom_headers)
        resp = self.put(url, data=data, headers=custom_headers)

        if all([self.is_nexus, pom, metadata]): self.update_metadata(params['g'], repo)

        return resp

    def update_metadata(self, gav=None, repo=None):
        """
        Update metadata in repo for gav
        :param gav: GAV
        :type gav: str
        :param repo: repository ID
        :type repo: str
        """
        if not self.is_nexus:
            return None

        if not repo: repo = self.__upload_repo
        if not repo: repo = os.getenv(self._env_prefix + self._env_upload_repo)
        if not repo: raise (ValueError("You must provide repo name"))

        _req = posixpath.join("service", "local", "metadata", "repositories", repo, 'content')
        
        if gav:
            _req = posixpath.join(_req, gav_to_path(gav, relaxed=True))

        return self.delete(_req)

    def __check_filter(self, str_artifact, str_filter, b_filter_revert):
        if not str_filter:
            return True

        re_filter = re.compile(r'' + str_filter)

        if (not b_filter_revert) and (re_filter.search(str_artifact) is not None):
            return True

        if (b_filter_revert) and (re_filter.search(str_artifact) is None):
            return True

        return False

    #TODO: replace 'filter' keyword argument - it is RESERVED in recent Python interpreters
    def ls(self, gav, repo=None, filter=None, filter_revert=False):
        """
        List of artifacts suitable for filter in repository given
        :param gav: gav to search for, may be regular expression
        :tytpe gav: str
        :param repo: repository ID
        :type repo: str
        :param filter: filter for artifacts
        :type filter: dict
        """
        if self.is_nexus: return self.__ls_nexus(gav, repo=repo, filter=filter, filter_revert=filter_revert)
        if self.is_artifactory: return self.__ls_artifactory(gav, repo=repo, filter=filter, filter_revert=filter_revert)
        raise NotImplementedError("Not implemented for '%s'" % posixpath.basename(self.root.rstrip(posixpath.sep)))

    # the same for specific systems
    def __ls_artifactory(self, gav, repo=None, filter=None, filter_revert=False):
        parsed_gav = parse_gav(gav, relaxed=True)
        search_parameters = parsed_gav.copy()

        # Artifactory does not provide search by packaging
        # so we have to filter it manually
        if "p" in search_parameters:
            del(search_parameters["p"])

        if repo:
            search_parameters['repos'] = repo

        artifacts = []

        _req = posixpath.join("api", "search", "gavc")

        # we need to return empty list if we found nothing
        try:
            resp = self.get(_req, params=search_parameters, headers={'X-Result-Detail': 'info'})
            resp = resp.json()
            resp = resp.get('results')
        except NexusAPIError as e:
            if e.code != 404:
                logging.exception(e)

            return artifacts

        if not resp:
            return artifacts


        for artifact in resp:
            # no path - no artifact

            if not artifact.get("path"):
                continue

            artifact_gav_dict = gav_from_path(artifact.get('path'))

            # historically we are not interested in POMs
            if artifact_gav_dict.get('p') == "pom":
                continue

            # if we are asked to search for exact packaging - filter it
            if all(["p" in parsed_gav, parsed_gav.get("p") != artifact_gav_dict.get("p")]):
                continue

            # check our artifact for filter
            generated_gav = gav_from_path(artifact.get('path'), dictionary=False)
            
            if self.__check_filter(generated_gav, filter, filter_revert):
                artifacts.append(generated_gav)

        return list(set(artifacts))

    def __ls_nexus(self, gav, repo=None, filter=None, filter_revert=False):
        g_s = parse_gav(gav, relaxed=True)
        g_p = dict()
        ls_artif = None

        if not g_s.get('a'):
            ls_artif = list()

        # Filter values of parsed GAV for Lucene Search search Nexus plugin
        # Packaging is to be excluded due to that plugin "feature"
        for ch_key in ['g', 'a', 'v', 'c']:
            if g_s.get(ch_key):
                g_p[ch_key] = g_s[ch_key]

        if repo:
            g_p['repositoryId'] = repo

        resp = self.get(posixpath.join("service", "local", "lucene", "search"), g_p)
        xml = ElementTree.fromstring(resp.content)
        artifacts = []
        for artifact in xml.find('data').findall('artifact'):
            for hit in artifact.find('artifactHits').findall('artifactHit'):
                b_ext_found = False
                if not 'p' in g_s:
                    b_ext_found = True; # do not search if we was not asked for it

                for ext_t in hit.find('artifactLinks').findall('artifactLink'):
                    ext = ext_t.find('extension')

                    if (ext is not None and ext.text == 'pom'):
                        continue

                    clsf = ext_t.find('classifier')
                    str_artifact_full = gav_to_str({
                        'g': artifact.find('groupId').text,
                        'a': artifact.find('artifactId').text,
                        'v': artifact.find('version').text,
                        'p': ext.text if ext is not None else None,
                        'c': clsf.text if clsf is not None else None})

                    __unused, str_artifact = str_artifact_full.split(':', 1)
                    
                    if not self.__check_filter(str_artifact, filter, filter_revert):
                        continue

                    if not g_s.get('p'):
                        if str_artifact_full not in artifacts \
                                and self.exists(str_artifact_full, repo=repo):
                            artifacts.append(str_artifact_full)

                        continue

                    if ext is not None and g_s.get('p') != ext.text:
                        continue

                    b_ext_found = bool(ext is not None and ext.text)

                    if str_artifact_full not in artifacts:
                        artifacts.append(str_artifact_full)

                if not b_ext_found:
                    # produce fake gav and check if exist
                    str_artifact_full = gav_to_str({
                        'g': artifact.find('groupId').text,
                        'a': artifact.find('artifactId').text,
                        'v': artifact.find('version').text,
                        'p': g_s.get('p'),
                        'c': g_s.get('c')})

                    if str_artifact_full in artifacts:
                        continue

                    __unused, str_artifact = str_artifact_full.split(':', 1)

                    if not self.__check_filter(str_artifact, filter, filter_revert):
                        continue

                    if self.exists(str_artifact_full):
                        artifacts.append(str_artifact_full)

            if ls_artif is not None and artifact.find('artifactId').text not in ls_artif:
                ls_artif.append(artifact.find('artifactId').text)

        if ls_artif is not None:
            for str_art in ls_artif:
                # here we may not use gav_to_str since
                # we need empty values for some attributes, 
                # but gav_to_str silently skips them
                # or raises undesirable exception
                str_gav_c = ':'.join([
                    g_s.get('g'), 
                    str_art,
                    g_s.get('v', ""),
                    g_s.get('p', ""),
                    g_s.get('c', "")])

                artifacts += self.__ls_nexus(str_gav_c, repo, filter, filter_revert)

        return list(set(artifacts))

    def info(self, gav, repo=None):
        """
        Get md5 checksum for a gav from file
        :param gav:  string, g:a:v:p
        :type gav: str
        :param repo: string, repository_id
        :type repo: str
        :return: artifact information: md5sum, mime-type
        :return type: dict, None
        """
        if not repo: repo = self.repo_default

        if self.is_nexus: return self.__info_nexus(gav, repo=repo)
        if self.is_artifactory: return self.__info_artifactory(gav, repo=repo)
        raise NotImplementedError("Not implemented for '%s'" % posixpath.basename(self.root.rstrip(posixpath.sep)))

    ### the same as general info for different systems
    def __info_artifactory(self, gav, repo):
        _req = posixpath.join("api", "storage", repo, gav_to_path(gav))

        try:
            http_resp = self.get(_req)
        except NexusAPIError as e:
            # Artifactory returns 404 if file is not found
            # or information about it is not (yet?) available
            if e.code != 404: 
                logging.exception(e)

            return None

        if not http_resp or not http_resp.content:
            return None

        response_json = http_resp.json()

        if response_json is None:
            return None

        # NOTE about 'dict()' default value for first 'get'
        # we do not want to raise an exception "NoneType object has no 'get' attribute"
        artifact_md5 = response_json.get('checksums', dict()).get('md5')
        artifact_mime = response_json.get('mimeType')

        if any([not artifact_md5, not artifact_mime]):
            return None

        return {"md5": artifact_md5, "mime": artifact_mime}

    def __info_nexus(self, gav, repo):
        str_path = posixpath.join("service", "local", "repositories", repo, "content", gav_to_path(gav))
        dict_parms = {"describe" : "info"}
        http_resp = self.get(str_path, dict_parms)

        if not http_resp or not http_resp.content:
            return None

        obj_xml = ElementTree.fromstring(http_resp.content)

        # NOTE: xml.etree.ElementTree drops warning if simple 'not obj_xml' is used
        #       explicitly comparison is strictly recommended
        if (obj_xml is None):
            return None

        xml_tag_md5 = obj_xml.find("./data/md5Hash")
        xml_tag_mime = obj_xml.find("./data/mimeType")

        if any([xml_tag_md5 is None, xml_tag_mime is None]):
            return None

        return {"md5" : xml_tag_md5.text, "mime" : xml_tag_mime.text}


    def remove(self, gav=None, repo=None):
        """
        Remove artifact by gav
        :param gav: string, artifact gav to remove
        :param repo: string, repository containing artifact
        :return: Response, response for delete request
        """

        if not repo:
            repo = self.__upload_repo
        if not repo:
            repo = os.getenv(self._env_prefix + self._env_upload_repo)
        if not repo:
            raise (ValueError("You must provide repo name"))

        _req = self.__gav_get_url(gav, repo=repo, relaxed=True)

        return self.delete(_req)




# requires python-requests rpm package
import requests
from .API import HttpAPI, HttpAPIError
from string import Template
import json
from collections import namedtuple, OrderedDict
import re
import xml.etree.ElementTree as ET;
import posixpath
import sys


class JenkinsError(HttpAPIError):
    def __str__(self):
        return 'Error communicating with Jenkins: ' + self.text + ': Code ' + str(self.code) + ' ' + self.url


class Jenkins(HttpAPI):
    """
    Simple and easy-to-use Jenkins interface

    >>> def test(port):
    ...     jenk = Jenkins('http://127.0.0.1:'+str(port))
    ...     jenk.disable_job('job.to_disable')
    ...     jenk.enable_job('job.to_enable')
    ...     jenk.delete_job('job.to_delete')
    ...     jenk.get_job('job.to_get')
    ...     jenk.put_job('job.to_put','this is a job')
    ...
    >>> from TestServer import test_wrapper
    >>> test_wrapper(test,6)
    "POST /job/job.to_disable/disable HTTP/1.1" 200 -
    "POST /job/job.to_enable/enable HTTP/1.1" 200 -
    "POST /job/job.to_delete/doDelete HTTP/1.1" 200 -
    "GET /job/job.to_get/config.xml/api/json HTTP/1.1" 200 -
    "GET /job/job.to_put/config.xml/api/json HTTP/1.1" 200 -
    "POST /job/job.to_put/config.xml HTTP/1.1" 200 -

    """

    _error = JenkinsError
    _env_prefix = 'JENKINS'

    def re(self, req):
        if req and req[0] == '!':
            req = req[1:]
            return posixpath.join( self.root, req );
        return posixpath.join( self.root, req, 'api', 'json' );

    def post(self, req, *vargs, **kvargs):
        # We need all post requests to be processed in a different way in re(),
        # but do not wan't to rewrite post(). So we pass a little flag to be checked in re()
        return super(Jenkins, self).post('!' + req, *vargs, **kvargs)

    def list_jobs(self, prefix=None, suffix=None):
        """Gets job list"""
        jobs = self.get('', params={'tree': 'jobs[name]'})
        return filter(lambda x: (not prefix or x.startswith(prefix)) and (not suffix or x.endswith(suffix)),
                      map(lambda x: x['name'], jobs.json().get('jobs', [])))
    def get_node(self, node):
        """Gets config.xml for specific slave"""
        config = self.get( posixpath.join( 'computer', node, 'config.xml') );
        return config.text

    def put_node(self, node, config):
        """Saves config.xml for a node (or creates new node)"""
        request = posixpath.join( 'computer', node, 'config.xml' );
        params = None
        headers = {'content-type' : 'application/xml'}

        try:
            self.get_node(node)
        except JenkinsError as e:
            if e.code == 404: request = posixpath.join( 'computer', 'doCreateItem' );
            params = {'name': node}

        if self.readonly: return None
        put = self.post(request, data=config, params=params, headers=headers)
        return put

    def delete_node( self, node ):
        """Deletes node from CI"""
        if self.readonly: return None
        return self.post( posixpath.join( 'computer', node, 'doDelete' ) );

    def get_node_label_from_config( self, config ):
        """
        Get the exact XML tag value with XPATH: label
        :param self: self reference
        :param config: xml string with a configuration of node
        :return: <label> tag text
        """
        obj_xml = ET.fromstring( config );
        xml_tag = obj_xml.find( 'label' );

        if ( xml_tag is None ) or ( xml_tag.text is None ) or ( len( xml_tag.text.strip() ) == 0 ):
            return "";

        return xml_tag.text;

    def get_node_label(self, node, config=None):
        if not config: config = self.get_node(node)
        return self.get_node_label_from_config( config ).split(' ')

    def set_node_label_in_config( self, config, label ):
        """
        set <label> tag in XML config
        :param self: self reference
        :param config: string with XML node configuration
        :return: string with edited 'config' with <lable> tag content set to 'label'
        """
        obj_xml = ET.fromstring( config );
        xml_tag = obj_xml.find( 'label' );

        if ( xml_tag is None ):
            xml_tag = ET.SubElement( obj_xml, 'label' );

        xml_tag.text = label;

        return ET.tostring( obj_xml, encoding='utf8' );

    def set_node_label(self, node, label):
        config = self.get_node(node)
        config = self.set_node_label_in_config( config, ' '.join(label))
        return self.put_node(node, config)

    def enable_node(self, job):
        if self.readonly: return None
        return self.post( posixpath.join( 'computer', job, 'enable') );

    def disable_node(self, job):
        if self.readonly: return None
        return self.post( posixpath.join( 'computer', job, 'disable') )

    def is_node_idle(self, node):
        """Retrieve 'idle' attribute of a node via newer JSON API (may require Jenkins 2.x)."""
        return json.loads(
            self.get("computer/{}/api/json".format(node)).content
        )['idle']

    def list_nodes( self ):
        """
        List all active nodes (except master one)
        :return: List of strings with node names, possible empty
        """
        raw_data = self.get( "computer", params = { "tree" : "computer[displayName]" } )
        str_data = raw_data.text

#        return map( lambda x: x[ "displayName" ], json.loads( self.get( "computer", params = { "tree" : "computer[displayName]" } ).text )[ "computer" ] );
        return [x[ "displayName" ] for x in json.loads( str_data )[ "computer" ]];


    def job_xml_enable( self, config ):
        """
        Enables jenkins job by modifying XML tag (XPATH=) in the configuration
        :param self: self reference
        :param config: xml string from a config
        :return: xml string with <disabled> tag value modified to 'true'
        """
        obj_xml = ET.fromstring( config );
        xml_tag = obj_xml.find( 'disabled' );

        if xml_tag is None:
            return config;

        xml_tag.text = "false";

        return ET.tostring( obj_xml, encoding='utf8' );

    def copy_job(self, config, j_name, parameters={}, enabled=True, use_template=False):
        """Create or update job by template"""

        if config[0] != '<':    config = self.get_job(config)
        parameters['name'] = j_name
        if enabled:
            config = self.job_xml_enable( config );

        cfg_type = str (type (config) )
        if cfg_type.find('byte') != -1:
            config = str (config, 'UTF-8')

        if use_template:
            config = Template(config).safe_substitute(parameters)
        else:
            config = config.format(**parameters)

        put = self.put_job(j_name, config)
        #		if enabled: self.enable_job(j_name)
        return put

    def get_job(self, job):
        """Gets config.xml for a job"""
        config = self.get( posixpath.join( 'job', job, 'config.xml') );
        return config.text

    def put_job(self, job, config):
        """Saves config.xml for a job (or creates new job)"""
        request = posixpath.join( 'job', job, 'config.xml' )
        params = None
        headers = {'content-type' : 'application/xml'}
        try:
            self.get_job(job)
        except JenkinsError as e:
            if e.code == 404: request = 'createItem'
            params = {'name': job}

        if self.readonly: return None
        put = self.post(request, data=config, params=params, headers=headers)
        return put

    def enable_job(self, job):
        if self.readonly: return None
        return self.post( posixpath.join( 'job', job, 'enable' ) );

    def disable_job(self, job):
        if self.readonly: return None
        return self.post( posixpath.join ( 'job', job, 'disable' ) );

    def delete_job(self, job):
        if self.readonly: return None
        return self.post( posixpath.join( 'job', job, 'doDelete' ) );

    def run_job(self, job, params=None):
        if self.readonly: return None
        if params == None:
            response=self.post( posixpath.join( 'job', job, 'build') );
        else:
            response=self.post( posixpath.join( 'job', job, 'buildWithParameters' ), data=params)

        queue_item=QueueItem.from_jenkins_response(response, self)
        return queue_item

    def list_builds( self, node = None ):
        """
        Get list of Builds which are currently build on the node given
        :param node: list only jobs building on the node specified
        :type node: string
        :return: dictionary with node_names as a key and list of jobs dictionaries building on the node as { "job_name": ... , "build_number" : ... }, list may be None
        """
        str_request = "";
        dict_parms = { "tree" : "jobs[name,lastBuild[building,fullDisplayName,number,url,builtOn]]" };

        if node is not None:
            str_request = posixpath.join( "computer", node );
            dict_parms[ "tree" ] = "executors[idle,currentExecutable[idle,building,builtOn,number,fullDisplayName,url]]";

        response = self.get( str_request, params = dict_parms );
        dict_resp = json.loads( response.text );

        if node is not None:
            if len( dict_resp[ "executors" ] ) == 0:
                return { node: list() };

            ls_jobs = map( lambda x: { "name" : None, "lastBuild" : ( None if x[ "idle" ] else x[ "currentExecutable" ] ) }, dict_resp[ "executors" ] );

        else :
            ls_jobs = dict_resp[ "jobs" ];


        dict_result = dict();

        for dict_job in ls_jobs:
            if "lastBuild" not in dict_job or dict_job[ "lastBuild" ] is None or not dict_job[ "lastBuild" ][ "building" ]:
                continue; # inactive build, out of interes

            str_job_name = dict_job[ "name" ];

            if not str_job_name:
                str_job_name = dict_job[ "lastBuild" ][ "url" ].split( posixpath.sep + "job" + posixpath.sep, 1 )[1].split( posixpath.sep, 1 )[0];

            str_builder_node = dict_job[ "lastBuild" ][ "builtOn" ];

            if str_builder_node not in dict_result:
                dict_result[ str_builder_node ] = list();

            dict_result[ str_builder_node ].append( { "job_name" : str_job_name, "build_nubmer" : dict_job[ "lastBuild" ][ "number" ] } );

        return dict_result;

    def list_queue( self ):
        """
        Get queue items, perhaps empty if no jobs queued
        :return: list of QueueItems, possible empty
        """
        ls_result = list();
        obj_resp = self.get( "queue" );
        dict_resp = json.loads( obj_resp.text or {} );

        if not 'items' in dict_resp:
            return ls_result;

        for dict_task in dict_resp[ 'items' ]:
            if not 'task' in dict_task:
                continue;

            if not 'name' in dict_task[ 'task' ]:
                continue;

            if not 'id' in dict_task:
                continue;

            ls_result.append( QueueItem( dict_task[ 'id' ], self, dict_task ) );
        
        return ls_result;

    def unqueue_item( self, int_id ):
        """
        remove item with ID given from queue
        :param int_id: id
        :type int_id: integer
        """
        if int_id is None or not isinstance( int_id, int ):
            raise TypeError( "Wrong type of arguments: int_id is not integer" );

        try:
            self.post( posixpath.join( 'queue','cancelItem'), data = {}, params = { "id": int_id } );
        except:
            pass; # since jenkins returns 404 code for this type of operation and it is normal


from enum import Enum
class BuildStatus(Enum):
    BUILDING=0
    SUCCESS=1
    FAILURE=2
    ABORTED=3
    UNSTABLE=4
    NOT_BUILD=5


class Build(object):
    """
    Class designating launched build with known number.
    """

    def __init__(self, job_name, build_number, jenkins_client):
        self.job_name=job_name
        self.build_number=build_number
        self.jenkins_client=jenkins_client
        self._build_status=BuildStatus.BUILDING

    def get_status(self):
        self._update_build_status()
        return self._build_status

    def _update_build_status(self):
        if self._build_status!=BuildStatus.BUILDING:
            return self._build_status
        build_path = posixpath.join( "job", self.job_name, str( self.build_number ), 'api', 'json' )
        response=self.jenkins_client.get(build_path)
        if response.status_code==404:
            raise JenkinsError("No build exist at %s" % build_path)
        parsed_response=json.loads(response.text or "{}")
        status_value=parsed_response["result"]
        if not status_value:
            self._build_status=BuildStatus.BUILDING
        else:
            if sys.version_info.major == 2:
                self._build_status = getattr(BuildStatus, status_value)
            else:
                self._build_status=BuildStatus[status_value]


class QueueItem(object):
    """ 
    Class used to hold information about build request.
    Initially it only knows its queue id, which can be used to get concrete build later.
    """

    @classmethod
    def from_jenkins_response(cls, response, jenkins_client):
        location_url=response.headers["Location"]
        search_result=re.search( r'' + posixpath.join( "queue", "item", "(?P<queue_id>\d+)" ), location_url)
        queue_id=int(search_result.group("queue_id"))
        return QueueItem(queue_id, jenkins_client)
    
    def __init__(self, queue_id, jenkins_client, dict_details = None):
        """
        :param resource: id in build queue (assigned at build launch request)
        :param required: jenkins.Jenkins instance
        """
        self.queue_id=queue_id
        self.jenkins_client=jenkins_client
        self._build=None
        self._task=None;

        if not dict_details:
            return;

        if "task" in dict_details:
            self._task = dict_details[ "task" ];

    @property
    def task( self ):
        """
        :return: dictionary with task parameters - job name, url, color - as it returned by Jenkins
        """
        if not self._task:
            self._try_update_queue_status();
        return self._task;

    @property
    def job_name( self ):
        """
        :return: job name of the queue item
        """

        if not self._task:
            self._try_update_queue_status();

        if not self._task:
            return None;

        if "url" not in self._task:
            self._try_update_queue_status();

        if "url" not in self._task:
            return None;

        return self._task[ "url" ].split( posixpath.sep + "job" + posixpath.sep, 1 )[1].split( posixpath.sep, 1 )[0];

    def is_running(self):
        self._try_update_queue_status()
        return bool(self._build)

    def get_build(self):
        self._try_update_queue_status()
        if self._build:
            return self._build
        else:
            raise BuildStatusException("Build is not started yet")

    def _try_update_queue_status(self):
        if self._build:
            return
        item_path= posixpath.join( "queue", "item", str( self.queue_id ), "api", "json" );
        response=self.jenkins_client.get(item_path)
        if response.status_code==404:
            raise JenkinsError("No queue item exist at %s" % item_path)
        parsed_response=json.loads(response.text or "{}")
        if "executable" in parsed_response:
            self._build=Build(parsed_response["task"]["name"],
                              parsed_response["executable"]["number"],
                              self.jenkins_client)

        if not self._task and "task" in parsed_response:
            self._task = parsed_response[ "task" ];
    

class BuildStatusException(Exception):
    pass


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import logging
import os
import re
import unittest

import sys
if sys.version_info >= (3, 3):
    from unittest.mock import patch
else:
    from mock import patch

from oc_cdtapi.DmsAPI import DmsAPI

import json

class FakeJson(object):
    def __init__(self):
        self.response = None

    def setresp(self, resp):
        self.response = resp

    def json(self):
        return self.response

class _Response(object):
    """
    Fake response object
    """

    def __init__(self, data):
        self.content = data
        self.status_code = 200

    def json(self):
        return json.loads(self.content)


class _Session(object):
    """
    Fake requests session
    """

    def __init__(self, handler):
        self.handler = handler

    def get(self, req, params = None, **other):
        r = req
        if params is not None and len(params) > 0: 
            r += '?' + '&'.join(map(lambda x,y: x+'='+y, list(params.items()))) # this makes from params dict a URL string
            return _Response(self.handler(r))
        else:
            components_json = FakeJson()
            components_json.setresp({"components":[{"id":"appserver"}]})
            return components_json


class _DmsAPI(DmsAPI):

    def __init__(self, *args, **argv):
        super(_DmsAPI, self).__init__(*args, **argv) # calling constructor in DmsAPI
        self.web = _Session(self._read_url) # here we redefine real requests session with fake one


    def get(self, url, params = None, headers = None, verify=False):
        req_a = '/'.join(url)
        if re.match('.+\/component\/.+\/version\/.+\/distribution\/.+\/gav$',req_a) and params: req_a = req_a + '.some'
        data = self._read_url(req_a)
        resp = FakeJson()
        resp.setresp(json.loads(data))
        return resp


    def _read_url(self,url):
        """
        override http request method
        """
        if re.match('.+\/component\/.+\/versions',url):
            return '{"versions":[{"version":"1.7.796","status":"RELEASE","versionInfo":{"snapshot":false,"rcVersion":false,"major":2,"minor":0,"service":1950,"itemsCount":4,"fix":0,"buildNumber":0}}]}'
        elif re.match('.+\/component\/.+\/version\/.+\/distribution\/list$',url):
            return '[{"name":"appserver","classifier":"rhel6-linux-x64"},{"name":"appserver","classifier":"rhel7-linux-x64"}]'
        elif re.match('.+\/component\/.+\/version\/.+\/distribution\/.+\/gav.+$',url):
            return '{"groupId":"com.localhost.distribution.appserver","artifactId":"appserver","version":"1.7.796","classifier":"windows-x64","packaging":"zip"}'
        elif re.match('.+\/component\/.+\/version\/.+\/distribution\/.+\/gav$',url):
            return '{"groupId":"com.localhost.distribution.appserver","artifactId":"appserver","version":"1.7.796","packaging":"zip"}'
        return '[]'



class TestDmsAPI(unittest.TestCase):


#    def test_dms(self):
#        """
#        Test if dms is properly configured and alive
#        """
#        da = _DmsAPI()
#        data = da.ping_dms()
#        self.assertRegexpMatches(data, '.+links+')
#
    @patch.dict(os.environ, {"DMS_URL": "file://dev/null/","DMS_TOKEN": "some-hashed-value", "DMS_CRS_URL": "file://dev/null/"})
    def setUp(self):
        self.dms_api = _DmsAPI()

    def test_get_components(self):
        test_ok = False
        all_c = self.dms_api.get_components()
        if {'id': 'appserver'} in all_c:
            test_ok = True
        self.assertTrue(test_ok)


    def test_get_versions(self):
        test_ok = False
        component = 'appserver'
        all_v = self.dms_api.get_versions(component)
        if '1.7.796' in all_v:
            test_ok = True
        self.assertTrue(test_ok)


    def test_get_types(self):
        test_ok = False
        all_t = self.dms_api.get_types()
        if 'distribution' in all_t:
            test_ok = True
        self.assertTrue(test_ok)


    def test_get_artifacts(self):
        test_ok = False
        component = 'appserver'
        ctype = 'distribution'
        version = '1.7.796'
        all_a = self.dms_api.get_artifacts(component, version, ctype)
        for a in all_a:
            if a['name'] == 'appserver':
                test_ok = True
        self.assertTrue(test_ok)


    def test_get_gav_no_classifier(self):
        test_ok = False
        component = 'appserver'
        ctype = 'distribution'
        version = '1.7.796'
        artifact = 'appserver'
        classifier = None
        gav = self.dms_api.get_gav(component, version, 'distribution', artifact, classifier)
        self.assertEqual(gav,'com.localhost.distribution.appserver:appserver:1.7.796:zip')


    def test_get_gav_classifier(self):
        test_ok = False
        component = 'appserver'
        ctype = 'distribution'
        version = '1.7.796'
        artifact = 'appserver'
        classifier = 'windows-x64'
        gav = self.dms_api.get_gav(component, version, 'distribution', artifact, classifier)
        self.assertEqual(gav,'com.localhost.distribution.appserver:appserver:1.7.796:zip:windows-x64')



if __name__ == '__main__':
    unittest.main()



import json
import logging
import os
import posixpath
import re
import sys

from oc_cdtapi.API import HttpAPI, HttpAPIError
from collections import namedtuple, defaultdict
from datetime import datetime, timedelta

if sys.version_info.major == 3:
    from urllib.parse import urljoin
else:
    from urlparse import urljoin

class ForemanAPIError(HttpAPIError):
    pass

class ForemanAPI(HttpAPI):
    """
    A simple client for Foreman's REST API
    """

    _error = ForemanAPIError
    _env_prefix = "FOREMAN"

    headers = {"Accept": "version=2,application/json", "Content-Type": "application/json"}

    def __init__(self):
        """
        Class initialization, setting the default values for a new host
        """
        HttpAPI.__init__(self)
        class_defaults = namedtuple("values", "exp_date location_id hostgroup deploy_on")
        exp_date = self._set_expiration()
        location_id = 5
        hostgroup = 11
        deploy_on = 1

        self.defs = class_defaults(exp_date, location_id, hostgroup, deploy_on)

        if os.environ.get ('FOREMAN_API_VERSION'):
            self.apiversion = int (os.environ.get ('FOREMAN_API_VERSION') )
        else:
            self.apiversion = 1

    
    def re(self, req):
        return urljoin (self.root, posixpath.join ("api", req) )


    def get_environment (self, env_name):
        """
        wrapper for api v1/v2
        """
        logging.debug ('Reached get_environment')
        logging.debug ('env_name = [%s]' % env_name)
        if self.apiversion == 1:
            logging.error ('Not supported in v1, returning None')
            return None
        elif self.apiversion == 2:
            logging.debug ('Passing to get_environment_v2')
            return self.get_environment_v2 (env_name)


    def get_environment_v2 (self, env_name):
        """
        returns environment id for env_name
        """
        logging.debug ('Reached get_environment_v2')
        logging.debug ('env_name = [%s]' % env_name)
        params = {'search': 'name=%s' % env_name}
        response = self.get ('environments', params=params).json ()
        results = response.get ('results')
        for result in results:
            if result.get ('name') == env_name:
                env_id = result.get ('id')
                logging.debug ('Found environment [%s]' % env_id)
                return env_id
        logging.error ('Could not find environment for [%s], returning None' % env_name)
        return None


    def get_owner (self, user_login):
        """
        wrapper for api v1/v2
        """
        logging.debug ('Reached get_owner')
        if self.apiversion == 1:
            logging.debug ('Passing to get_owner_v1')
            return self.get_owner_v1 (user_login)
        elif self.apiversion == 2:
            logging.debug ('Passing to get_owner_v2')
            return self.get_owner_v2 (user_login)


    def get_owner_v1 (self, user_login):
        """
        Looks for user id in the Foreman DB
        """
        logging.debug ('Reached get_owner_v1')
        try:
            params = {'search': 'login=%s' % user_login}
            response = self.get('users', params=params)
        except ForemanAPIError as err:
            raise(err)

        data = json.loads(response.text)

        try:
            user_id = data["results"][0]["id"]
        except IndexError:
            user_id = None

        return user_id


    def get_owner_v2 (self, user_login):
        """
        did not change since v1
        """
        logging.debug ('Reached get_owner_v2')
        logging.debug ('Passing to get_owner_v1')
        return self.get_owner_v1 (user_login)


    def get_usergroup_id (self, group_name):
        """
        wrapper for api v1/v2
        """
        logging.debug ('Reached get_usergroup_id')
        if self.apiversion == 1:
            logging.debug ('Passing to get_usergroup_id_v1')
            return self.get_usergroup_id_v1 (group_name)
        if self.apiversion == 2:
            logging.debug ('Passing to get_usergroup_id_v2')
            return self.get_usergroup_id_v2 (group_name)


    def get_usergroup_id_v1 (self, group_name):
        """
        Looks for usergroup id in the Foreman DB
        """
        logging.debug ('Reached get_usergroup_id_v1')
        if re.search("\s", group_name):
            group_name = "%22{}%22".format(re.sub("\s", "%20", group_name))
        try:
            params = {'search': 'name=%s' % group_name}
            response = self.get('usergroups', params=params)
        except ForemanAPIError as err:
            raise(err)

        data = response.json()
        
        try:
            group_id = data["results"][0]["id"]
        except IndexError:
            group_id = None

        return group_id


    def get_usergroup_id_v2 (self, group_name):
        """
        Looks for usergroup id in the Foreman DB
        """
        logging.debug ('Reached get_usergroup_id_v2')
        logging.debug ('Passing to get_usergroup_id_v1')
        return self.get_usergroup_id_v1 (group_name)


    def _set_expiration(self):
        """
        A private method which sets the default expiration date (1 year from the current date)
        """
        logging.debug ('Reached _set_expiration')
        return str((datetime.now() + timedelta(days=90)).strftime('%d/%m/%Y'))


    def create_host (self, hostname = None, cores = 1, memory = 4096, disk = 50, owner_id = None,
                    exp_date = None, location_id = None, hostgroup = None,
                    deploy_on = None, custom_json = None):
        """
        wrapper for api v1/v2
        """
        logging.debug ('Reached create_host')
        logging.debug ('hostname = [%s]' % hostname)
        logging.debug ('cores = [%s]' % cores)
        logging.debug ('memory = [%s]' % memory)
        if self.apiversion == 1:
            logging.debug ('Passing to create_host_v1')
            self.create_host_v1 (hostname, cores, memory, disk, owner_id,
                    exp_date, location_id, hostgroup,
                    deploy_on, custom_json)
        elif self.apiversion == 2:
            logging.debug ('Passing to create_host_v2')
            self.create_host_v2 (hostname, custom_json)


    def create_host_v1 (self, hostname, cores, memory, disk, owner_id,
                    exp_date, location_id, hostgroup,
                    deploy_on, custom_json):
        """
        Creates a host using the default parameters or the ones from an external json
        note that create_vm in engine actually sends db_task instead of hostname and custom_json, other parms are ignored
        """
        logging.debug ('Reached create_host_v1')
        logging.debug ('hostname = [%s]' % hostname)

        if not exp_date: exp_date = self.defs.exp_date
        if not location_id: location_id = self.defs.location_id
        if not hostgroup: hostgroup = self.defs.hostgroup
        if not deploy_on: deploy_on = self.defs.deploy_on

        default_params = {
            "name": hostname,
            "compute_attributes": {
                "start": "1",
                "cpus": cores,
                "memory_mb": memory,
                "volumes_attributes": {
                    0: {
                        "size_gb": disk
                    },
                },
            },
            "location_id": location_id,
            "hostgroup_id": hostgroup,
            "compute_resource_id": deploy_on,
            "is_owned_by": owner_id,
            "build": "true",
            "expired_on": exp_date
        }

        if custom_json:
            custom = json.loads(custom_json)
            default_params.update(custom)

        if not default_params["is_owned_by"]:
            raise ForemanAPIError("The owner id is not specified")

        if not default_params["name"]:
            raise ForemanAPIError("The hostname is not specified")

        logging.debug ("ForemanAPI is about to send the following payload:")
        logging.debug (default_params)
        request = self.post("hosts", headers = self.headers, json = default_params)


    def create_host_v2 (self, task, custom_json):

        logging.debug ('Reached create_host_v2')
        logging.debug ('task = [%s]' % task)
        hostname = task ['task_content'] ['resources'] ['name']
        cores = 1
        memory = 4096
        disk = 50
        owner_id = None

        exp_date = self.defs.exp_date
        location_id = self.defs.location_id
        hostgroup = self.defs.hostgroup
        deploy_on = self.defs.deploy_on
        domain_id = self.get_domain_id (hostname)
        arch_id = self.get_architecture_id ('x86_64')
        os_id = self.get_os_id ('CentOS Linux 7.9.2009')
        ptable_id = self.get_ptable_id (os_id, 'CDT LVM')

        default_params = {
            "name": hostname,
            "compute_attributes": {
                "start": "1",
                "cpus": cores,
                "memory_mb": memory,
                "volumes_attributes": {
                    0: {
                        "size_gb": disk
                    },
                },
            },
            "location_id": location_id,
            "hostgroup_id": hostgroup,
            "compute_resource_id": deploy_on,
            "is_owned_by": owner_id,
            "build": "true",
            "expired_on": exp_date,
            "domain_id": domain_id,
            "architecture_id": arch_id,
            "operatingsystem_id": os_id,
            "ptable_id": ptable_id
        }

        if custom_json:
            custom = json.loads(custom_json)
            default_params.update(custom)

        if not default_params["is_owned_by"]:
            raise ForemanAPIError("The owner id is not specified")

        if not default_params["name"]:
            raise ForemanAPIError("The hostname is not specified")

        if not default_params.get ('hostgroup_id'):
            hostgroup = self.get_hostgroup_id ('stands')
            logging.debug ('houstgroup_id is not set, setting default [%s]' % hostgroup)
            default_params ['hostgroup_id'] = hostgroup

        if not default_params.get ('environment_id'):
            env_id = self.get_environment ('development')
            logging.debug ('environment_id is not set, setting default [%s]' % env_id)
            default_params ['environment_id'] = env_id

        logging.debug ("ForemanAPI is about to send the following payload:")
        logging.debug (default_params)
        request = self.post("hosts", headers = self.headers, json = default_params)


    def get_architecture_id (self, arch_name):
        """
        wrapper for api v1/v2
        """
        logging.debug ('Reached get_architecture_id')
        logging.debug ('arch_name = [%s]')
        if self.apiversion == 1:
            logging.error ('Not supported in v1, returning None')
            return None
        elif self.apiversion == 2:
            logging.debug ('Passing to get_architecture_id_v2')
            return self.get_architecture_id_v2 (arch_name)


    def get_architecture_id_v2 (self, arch_name):
        """
        returns architecture id
        """
        logging.debug ('Reached get_architecture_id_v2')
        logging.debug ('arch_name = [%s]' % arch_name)
        params = {'search': 'name=%s' % arch_name}
        response = self.get ('architectures', params=params).json ()
        logging.debug ('Received response:')
        logging.debug (response)
        results = response.get ('results')
        for result in results:
            if result.get ('name') == arch_name:
                arch_id = result.get ('id')
                logging.debug ('Found architecture, id = [%s]' % id)
                return arch_id
        logging.error ('No architecture found, returning None')
        return None
      

    def get_domain_id (self, hostname):
        """
        wrapper for api v1/v2
        """
        logging.debug ('Reached get_domain_id')
        if self.apiversion == 1:
            logging.error ('Not supported in v1, returning None')
            return None
        elif self.apiversion == 2:
            logging.debug ('Passing to get_domain_id_v2')
            return self.get_domain_id_v2 (hostname)


    def get_domain_id_v2 (self, hostname):
        """
        Returns domain id if found
        """
        logging.debug ('Reached get_domain_id_v2')
        logging.debug ('hostname = [%s]' % hostname)
        domain = '.'.join (hostname.split ('.') [1:] )
        response = self.get ('domains/')
        j = response.json ()
        domains = j ['results']
        logging.debug ('Searching domain [%s]' % domain)
        logging.debug ('Domains list:')
        logging.debug (domains)
        for d in domains:
            if d.get ('name') == domain:
                return d.get ('id')
        logging.error ('Cannot find domain for host [%s]' % hostname)
        return None


    def get_host_info (self, hostname):
        """
        wrapper for api v1/v2
        """
        logging.debug ('Reached get_host_info')
        if self.apiversion == 1:
            logging.debug ('Passing to get_host_info_v1')
            return self.get_host_info_v1 (hostname)
        elif self.apiversion == 2:
            logging.debug ('Passing to get_host_info_v2')
            return self.get_host_info_v2 (hostname)


    def get_host_info_v1 (self, hostname):
        """
        Gathers all information about the host and returns it as a json
        """
        logging.debug ('Reached get_host_info_v1')
        logging.debug ('hostname = [%s]' % hostname)
        response = self.get("hosts/" + hostname)
        return response.json()


    def get_host_info_v2 (self, hostname):
        logging.debug ('Reached get_host_info_v2')
        logging.debug ('hostname = [%s]' % hostname)
        logging.debug ('passing to get_host_info_v1')
        return self.get_host_info_v1 (hostname)


    def get_ptable_id (self, os_id, ptable_name):
        """
        wrapper for api v1/v2
        """
        logging.debug ('Reached get_ptable_id')
        if self.apiversion == 1:
            logging.error ('Not supported in v1, returning None')
            return None
        elif self.apiversion == 2:
            logging.debug ('Passing to get_ptable_id_v2')
            return self.get_ptable_id_v2 (os_id, ptable_name)


    def get_ptable_id_v2 (self, os_id, ptable_name):
        """
        returns partition table id
        """
        logging.debug ('Reached get_ptable_id_v2')
        logging.debug ('os_id = [%s]' % os_id )
        logging.debug ('ptable_name = [%s]' % ptable_name)
        response = self.get ('operatingsystems/%s/ptables' % os_id).json ()
        logging.debug ('response is:')
        logging.debug (response)
        results = response.get ('results')
        for result in results:
            if result.get ('name') == ptable_name:
                ptable_id = result.get ('id')
                logging.debug ('Found ptable id = [%s]' % ptable_id)
                return ptable_id
        logging.error ('No ptable found, returning None')
        return None


    def update_host (self, hostname, payload):
        """
        wrapper for api v1/v2
        """
        logging.debug ('Reached update_host')
        if self.apiversion == 1:
            logging.debug ('Passing to update_host_v1')
            self.update_host_v1 (hostname, payload)
        elif self.apiversion == 2:
            logging.debug ('Passing to update_host_v2')
            self.update_host_v2 (hostname, payload)


    def update_host_v1 (self, hostname, payload):
        """
        Updates the host using the payload
        """
        logging.debug ('Reached update_host_v1')
        request = self.put("hosts/" + hostname, headers = self.headers, json = payload)


    def update_host_v2 (self, hostname, payload):
        """
        Updates the host using the payload
        """
        logging.debug ('Reached update_host_v2')
        logging.debug ('Passing to update_host_v1')
        self.update_host_v1 (hostname, payload)


    def delete_host (self, hostname):
        """
        wrapper for api v1/v2
        """
        logging.debug ('Reached delete_host')
        if self.apiversion == 1:
            logging.debug ('Passing to delete_host_v1')
            self.delete_host_v1 (hostname)
        elif self.apiversion == 2:
            logging.debug ('Passing to delete_host_v2')
            self.delete_host_v2 (hostname)


    def delete_host_v1 (self, hostname):
        """
        Deletes the specified host
        """
        logging.debug ('Reached delete_host_v1')
        request = self.delete("hosts/" + hostname, headers = self.headers)


    def delete_host_v2 (self, hostname):
        """
        Deletes the specified host
        """
        logging.debug ('Reached delete_host_v2')
        logging.debug ('Passing to delete_host_v1')
        self.delete_host_v1 (hostname)


    def puppet_class_info (self, classname):
        """
        wrapper for api v1/v2
        """
        logging.debug ('Reached puppet_class_info')
        if self.apiversion == 1:
            logging.debug ('Passing to puppet_class_info_v1')
            return self.puppet_class_info_v1 (classname)
        elif self.apiversion == 2:
            logging.debug ('Passing to puppet_class_info_v2')
            return self.puppet_class_info_v1 (classname)


    def puppet_class_info_v1 (self, classname):
        """
        Returns puppet class info
        """
        logging.debug ('Reached puppet_class_info_v1')
        response = self.get("puppetclasses/" + classname, headers = self.headers)
        return response.json()


    def puppet_class_info_v2 (self, classname):
        """
        Returns puppet class info
        """
        logging.debug ('Reached puppet_class_info_v2')
        logging.debug ('Passing to puppet_class_info_v1')
        return self.puppet_class_info_v1 (classname)


    def smart_class_info (self, scid):
        """
        wrapper for api v1/v2
        """
        logging.debug ('Reached smart_class_info')
        if self.apiversion == 1:
            logging.debug ('Passing to smart_class_info_v1')
            return self.smart_class_info_v1 (scid)
        if self.apiversion == 2:
            logging.debug ('Passing to smart_class_info_v2')
            return self.smart_class_info_v2 (scid)


    def smart_class_info_v1 (self, scid):
        """
        Returns smart class info
        """
        logging.debug ('Reached smart_class_info_v1')
        response = self.get("smart_class_parameters/" + str(scid), headers = self.headers)
        return response.json()


    def smart_class_info_v2 (self, scid):
        """
        Returns smart class info
        """
        logging.debug ('Reached smart_class_info_v2')
        logging.debug ('Passing to smart_class_info_v1')
        return self.smart_class_info_v1 (scid)


    def override_smart_class (self, scid, params):
        """
        wrapper for api v1/v2
        """
        logging.debug ('Reached mverride_smart_class')
        if self.apiversion == 1:
            logging.debug ('Passing to override_smart_class_v1')
            self.override_smart_class_v1 (scid, params)
        elif self.apiversion == 2:
            logging.debug ('Passing to override_smart_class_v2')
            self.override_smart_class_v1 (scid, params)


    def override_smart_class_v1 (self, scid, params):
        """
        Overrides smart class parameters
        """
        logging.debug ('Reached override_smart_class_v1')
        request = self.post("smart_class_parameters/" + str(scid) + "/override_values", headers = self.headers, data = params)


    def override_smart_class_v2 (self, scid, params):
        """
        Overrides smart class parameters
        """
        logging.debug ('Reached override_smart_class_v2')
        logging.debug ('Passing to override_smart_class_v1')
        self.override_smart_class_v1 (scid, params)


    def get_hostgroup_puppetclasses (self, hostgroup_id):
        """
        wrapper for api v1/v2
        """
        logging.debug ('Reached get_hostgroup_puppetclasses')
        if self.apiversion == 1:
            logging.debug ('Passing to get_hostgroup_puppetclasses_v1')
            return self.get_hostgroup_puppetclasses_v1 (hostgroup_id)
        elif self.apiversion == 2:
            logging.debug ('Passing to get_hostgroup_puppetclasses_v2')
            return self.get_hostgroup_puppetclasses_v2 (hostgroup_id)


    def get_hostgroup_puppetclasses_v1 (self, hostgroup_id):
        """
        Returns info about all hostgroup's puppetclasses
        """
        logging.debug ('Reached get_hostgroup_puppetclasses_v1')
        response = self.get("hostgroups/" + str(hostgroup_id) + "/puppetclasses", headers = self.headers)
        return response.json()


    def get_hostgroup_puppetclasses_v2 (self, hostgroup_id):
        """
        Returns info about all hostgroup's puppetclasses
        """
        logging.debug ('Reached get_hostgroup_puppetclasses_v2')
        params = {'hostgroup_id': str (hostgroup_id) }
        response = self.get ('puppetclasses/', params=params)
        return response.json ()


    def add_puppet_class_to_host (self, hostname, params):
        """
        wrapper for api v1/v2
        """
        logging.debug ('Reached add_puppet_class_to_host')
        if self.apiversion == 1:
            logging.debug ('Passing to add_puppet_class_to_host_v1')
            self.add_puppet_class_to_host_v1 (hostname, params)
        if self.apiversion == 2:
            logging.debug ('Passing to add_puppet_class_to_host_v2')
            self.add_puppet_class_to_host_v2 (hostname, params)


    def add_puppet_class_to_host_v1 (self, hostname, params):
        """
        Adds the required puppet class to the host
        """
        logging.debug ('Reached add_puppet_class_to_host_v1')
        request = self.post("hosts/" + hostname + "/puppetclass_ids", headers = self.headers, data = params)


    def add_puppet_class_to_host_v2 (self, hostname, params):
        """
        Adds the required puppet class to the host
        """
        logging.debug ('Reached add_puppet_class_to_host_v2')
        logging.debug ('Passing to add_puppet_class_to_host_v1')
        self.add_puppet_class_to_host_v1 (hostname, params)


    def get_subnets (self):
        """
        wrapper for api v1/v2
        """
        logging.debug ('Reached get_subnets')
        if self.apiversion == 1:
            logging.debug ('Passing to get_subnets_v1')
            return self.get_subnets_v1 ()
        elif self.apiversion == 2:
            logging.debug ('Passing to get_subnets_v2')
            return self.get_subnets_v2 ()


    def get_subnets_v1 (self):
        """
        Returns all available subnets
        """
        logging.debug ('Reached get_subnets_v1')
        subnets_count = self.get("subnets", headers = self.headers).json()["total"]
        response = self.get("subnets?per_page={}".format(subnets_count), headers = self.headers)
        return response.json()


    def get_subnets_v2 (self):
        """
        Return all available subnets
        """
        logging.debug ('Reached get_subnets_v2')
        logging.debug ('Passing to get_subnets_v1')
        return self.get_subnets_v1 ()


    def get_host_reports (self, hostname, last=False):
        """
        wrapper for api v1/v2
        """
        logging.debug ('Reached get_host_reports')
        if self.apiversion == 1:
            logging.debug ('Passing to get_host_reports_v1')
            return self.get_host_reports_v1 (hostname, last)
        elif self.apiversion == 2:
            logging.debug ('Passing to get_host_reports_v2')
            return self.get_host_reports_v2 (hostname, last)


    def get_host_reports_v1 (self, hostname, last):
        """
        Returns all reports (or only the last one) for the host
        """
        if not last:
            response = self.get("hosts/" + hostname + "/config_reports", headers = self.headers)
        else:
            response = self.get("hosts/" + hostname + "/config_reports/last", headers = self.headers)
        return response.json()


    def get_host_reports_v2 (self, hostname, last):
        """
        Returns all reports (or only the last one) for the host
        """
        logging.debug ('Reached get_host_reports_v2')
        logging.debug ('hostname = [%s]' % hostname)
        logging.debug ('last = [%s]' % last)
        if not last:
            params = {'search': 'host=%s' % hostname}
            response = self.get ('config_reports', params=params).json ()
            logging.debug ('Received response:')
            logging.debug (response)
        else:
            host_id = self.get_host_info (hostname).get ('id')
            logging.debug ('host_id = [%s]' % host_id)
            response = self.get ('hosts/%s/config_reports/last' % host_id).json ()
            logging.debug ('Received response:')
            logging.debug (response)
        return response


    def host_power (self, hostname, action):
        """
        wrapper for api v1/v2
        NB currently (2023-02-08) host power is managed via vsphere api
        """
        logging.debug ('Reached host_power')
        if self.apiversion == 1:
            logging.debug ('Passing to host_power_v1')
            self.host_power_v1 (hostname, action)
        elif self.apiversion == 2:
            logging.debug ('Passing to host_power_v2')
            self.host_power_v2 (hostname, action)


    def host_power_v1 (self, hostname, action):
        """
        Turns on/off power on the host
        """
        logging.debug ('Reached host_power_v1')
        actions = ["start", "stop"]
        if action not in actions:
            raise ForemanAPIError("500 - Incorrect power action was provided")
        params = json.dumps({"power_action": action})
        request = self.put("hosts/" + hostname + "/power", headers = self.headers, data = params)


    def host_power_v2 (self, hostname, action):
        """
        Turns on/off power on the host
        """
        logging.debug ('Reached host_power_v2')
        logging.debug ('Passing to host_power_v1')
        logging.debug ('host = [%s]' % host)
        logging.debug ('action = [%s]' % host)
        self.host_power_v1 (hostname, action) 


    def get_report (self, id):
        """
        wrapper for api v1/v2
        """
        logging.debug ('Reached get_report')
        if self.apiversion == 1:
            logging.debug ('Passing to get_report_v1')
            return self.get_report_v1 (id)
        elif self.apiversion ==  2:
            logging.debug ('Passing to get_report_v2')
            return self.get_report_v2 (id)


    def get_report_v1 (self, id):
        """
        Returns a foreman report with specified id
        :param id: int, id of the required report
        :return: response in json format
        """
        response = self.get("config_reports/" + str(id), headers = self.headers)
        return response.json()


    def get_report_v2 (self, id):
        """
        Returns a foreman report with specified id
        :param id: int, id of the required report
        :return: response in json format
        """
        logging.debug ('Reached get_report_v2')
        logging.debug ('Passing to get_report_v1')
        return self.get_report_v1 (id)

    
    def get_hostgroup_id (self, hostgroup_name):
        """
        wrapper for api v1/v2
        """
        logging.debug ('Reached get_hostgroup_id')
        if self.apiversion == 1:
            logging.debug ('Passing to get_hostgroup_id_v1')
            return self.get_hostgroup_id_v1 (hostgroup_name)
        elif self.apiversion == 2:
            logging.debug ('Passing to get_hostgroup_id_v2')
            return self.get_hostgroup_id_v2 (hostgroup_name)


    def get_hostgroup_id_v1 (self, hostgroup_name):
        """
        Returns id of the required hostgroup
        :param hostgroup_name: str
        :return: int
        """
        logging.debug ('Reached get_hostgroup_id_v1')
        hostgroups = self.get("hostgroups", headers = self.headers).json()["results"]
        for hostgroup in hostgroups:
            if hostgroup["name"] == hostgroup_name:
                return hostgroup.get ('id')


    def get_hostgroup_id_v2 (self, hostgroup_name):
        """
        Returns id of the required hostgroup
        :param hostgroup_name: str
        :return: int
        """
        logging.debug ('Reached get_hostgroup_id_v2')
        logging.debug ('Passing to get_hostgroup_id_v1')
        return self.get_hostgroup_id_v1 (hostgroup_name)


    def get_organization_id (self, organization_name):
        """
        wrapper for api v1/v2
        """
        logging.debug ('Reached get_organization_id')
        if self.apiversion == 1:
            logging.debug ('Passing to get_organization_id_v1')
            return self.get_organization_id_v1 (organization_name)
        elif self.apiversion == 2:
            logging.debug ('Passing to get_organization_id_v2')
            return self.get_organization_id_v2 (organization_name)


    def get_organization_id_v1 (self, organization_name):
        """
        Returns id of the required organization
        :param organization_name: str
        :return: int
        """
        logging.debug ('Reached get_organization_id_v1')
        organization_id = None
        organizations_count = self.get("organizations", headers = self.headers).json()["total"]
        organizations = self.get("organizations?per_page={}".format(organizations_count), headers = self.headers).json()["results"]
        
        try:
            organization_id = next(organization["id"] for organization in organizations if organization["name"] == organization_name)
            return organization_id
        except StopIteration:
            return None


    def get_organization_id_v2 (self, organization_name):
        """
        Returns id of the required organization
        :param organization_name: str
        :return: int
        """
        logging.debug ('Reached get_organization_id_v2')
        logging.debug ('Passing to get_organization_id_v1')
        return self.get_organization_id_v1 (organization_name)


    def get_os_id (self, os_name):
        """
        Wrapper for api v1/v2
        """
        logging.debug ('Reached get_os_id')
        logging.debug ('os_name = [%s]')
        if self.apiversion == 1:
            logging.error ('Not supported in v1, returning None')
            return None
        elif self.apiversion == 2:
            logging.debug ('Passing to get_os_id_v2')
            return self.get_os_id_v2 (os_name)


    def get_os_id_v2 (self, os_name):
        """
        returns operating system id
        """
        logging.debug ('Reached get_os_id_v2')
        logging.debug ('os_name = [%s]' % os_name)
        params = {'search': 'name=%s' % os_name}
        response = self.get ('operatingsystems', params=params).json ()
        logging.debug ('response is:')
        logging.debug (response)
        results = response.get ('results')
        for result in results:
            if result.get ('description') == os_name:
                os_id = result.get ('id')
                logging.debug ('Found os, id = [%s]' % os_id)
                return os_id
        logging.error ('OS not found, returning None')
        return None


    def set_host_expiry (self, hostname, expiry):
        """
        wrapper for api v1/v2
        """
        logging.debug ('Reached set_host_expiry')
        if self.apiversion == 1:
            logging.debug ('Passing to set_host_expiry_v1')
            self.set_host_expiry_v1 (hostname, expiry)
        elif self.apiversion == 2:
            logging.debug ('Passing to set_host_expiry_v2')
            self.set_host_expiry_v2 (hostname, expiry)


    def set_host_expiry_v1 (self, hostname, expiry):
        """
        Attempts to set host expiry date
        :param hostname: full hostname 
        :param expiry: expiry date in format yyyy-mm-dd
        """
        logging.debug ('Reached set_host_expiry_v1')
        pl={}
        pl['host']={}
        pl['host']['expired_on']=expiry
        try:
            self.update_host(hostname,pl)
        except ForemanAPIError as err:
            raise(err)
    

    def set_host_expiry_v2 (self, hostname, expiry):
        """
        Attempts to set host expiry date
        :param hostname: full hostname 
        :param expiry: expiry date in format yyyy-mm-dd
        """
        logging.debug ('Reached set_host_expiry_v2')
        logging.debug ('Passing to set_host_expiry_v1')
        self.set_host_expiry_v1 (hostname, expiry)


    def get_image_uuid (self, os_name, image_name):
        """
        wrapper api v1/v2
        """
        logging.debug ('Reached get_image_uuid')
        if self.apiversion == 1:
            logging.debug ('Passing to get_image_uuid_v1')
            return self.get_image_uuid_v1 (os_name, image_name)
        elif self.apiversion == 2:
            logging.debug ('Passing to get_image_uuid_v2')
            return self.get_image_uuid_v2 (os_name, image_name)


    def get_image_uuid_v1 (self, os_name, image_name):
        """
        Returns OS image uuid
        :param os_name: str
        :param image_name: str
        :return: str
        """
        logging.debug ('Reached get_image_uuid_v1')
        os_list = self.get("operatingsystems", headers=self.headers).json()["results"]
        try:
            os_id = next(os["id"] for os in os_list if os["description"] == os_name)
            images_list = self.get("operatingsystems/{}/images".format(os_id)).json()["results"]
            image_uuid = next(image["uuid"] for image in images_list if image["name"] == image_name)
            return image_uuid
        except StopIteration:
            return None

    
    def get_image_uuid_v2 (self, os_name, image_name):
        """
        Returns OS image uuid
        :param os_name: str
        :param image_name: str
        :return: str
        """
        logging.debug ('Reached get_image_uuid_v2')
        logging.debug ('Passing to get_image_uuid_v1')
        return self.get_image_uuid_v1 (os_name, image_name)


    def get_flavor_id (self, compute_resource_id, flavor_name):
        """
        wrapper api v1/v2
        """
        logging.debug ('Reached get_flavor_id')
        if self.apiversion == 1:
            logging.debug ('Passing to get_flavor_id_v1')
            return self.get_flavor_id_v1 (compute_resource_id, flavor_name)
        elif self.apiversion == 2:
            logging.debug ('Passing to get_flavor_id_v2')
            return self.get_flavor_id_v2 (compute_resource_id, flavor_name)


    def get_flavor_id_v1 (self, compute_resource_id, flavor_name):
        """
        :param compute_resource_id: int
        :param flavor_name: str
        :return: int
        """
        logging.debug ('Reached get_flavor_id_v1')
        flavors_list = self.get("compute_resources/{}/available_flavors".format(compute_resource_id)).json()["results"]
        try:
            flavor_id = next(flavor["id"] for flavor in flavors_list if flavor["name"] == flavor_name)
            return flavor_id
        except StopIteration:
            return None
    

    def get_flavor_id_v2 (self, compute_resource_id, flavor_name):
        """
        :param compute_resource_id: int
        :param flavor_name: str
        :return: int
        """
        logging.debug ('Reached get_flavor_id_v2')
        logging.debug ('Passing to get_flavor_id_v1')
        return self.get_flavor_id_v1 (compute_resource_id, flavor_name)


    def get_tenant_id (self, compute_resource_id):
        """
        wrapper api v1/v2
        """
        logging.debug ('Reached get_tenant_id')
        if self.apiversion == 1:
            logging.debug ('Passing to get_tenant_id_v1')
            return self.get_tenant_id_v1 (compute_resource_id)
        elif self.apiversion == 2:
            logging.debug ('Passing to get_tenant_id_v2')
            return self.get_tenant_id_v2 (compute_resource_id)


    def get_tenant_id_v1 (self, compute_resource_id):
        """
        :param compute_resource_id: int
        :return: str
        """
        logging.debug ('Reached get_tenant_id_v1')
        compute_resource_info = self.get("compute_resources/{}".format(compute_resource_id)).json()
        try:
            tenant_id = compute_resource_info["compute_attributes"][0]["attributes"]["tenant_id"]
            return tenant_id
        except KeyError:
            return None
    

    def get_tenant_id_v2 (self, compute_resource_id):
        """
        :param compute_resource_id: int
        :return: str
        """
        logging.debug ('Reached get_tenant_id_v2')
        logging.debug ('Passing to get_tenant_id_v1')
        return self.get_tenant_id_v1 (compute_resource_id)


    def get_hosts_uuids (self, hostnames):
        """
        wrapper api v1/v2
        """
        logging.debug ('Reached get_hosts_uuids')
        if self.apiversion == 1:
            logging.debug ('Passing to get_hosts_uuids_v1')
            return self.get_hosts_uuids_v1 (hostnames)
        elif self.apiversion == 2:
            logging.debug ('Passing to get_hosts_uuids_v2')
            return self.get_hosts_uuids_v2 (hostnames)


    def get_hosts_uuids_v1 (self, hostnames):
        """
        :param hostnames: list
        :return: dict
        """
        logging.debug ('Reached get_hosts_uuids_v1')
        hosts_total_qty = self.get("hosts", params={"per_page": 1}).json()["total"]
        hostnames_uuids = {host["name"]: host["uuid"] for host in self.get("hosts", params={"per_page": hosts_total_qty}).json()["results"]}
        uuids = {hostname: uuid for hostname, uuid in hostnames_uuids.items() if hostname in hostnames}
        return uuids
    

    def get_hosts_uuids_v2 (self, hostnames):
        """
        :param hostnames: list
        :return: dict
        """
        logging.debug ('Reached get_hosts_uuids_v2')
        logging.debug ('Passing to get_hosts_uuids_v1')
        return self.get_hosts_uuids_v1 (hostnames)


    def get_host_uuid (self, hostname):
        """
        wrapper api v1/v2
        """
        logging.debug ('Reached get_host_uuid')
        if self.apiversion == 1:
            logging.debug ('Passing to get_host_uuid_v1')
            return self.get_host_uuid_v1 (hostname)
        elif self.apiversion == 2:
            logging.debug ('Passing to get_host_uuid_v2')
            return self.get_host_uuid_v2 (hostname)


    def get_host_uuid_v1 (self, hostname):
        """
        :param hostname: str
        :return: str
        """
        logging.debug ('Reached get_host_uuid_v1')
        try:
            return self.get("hosts/{}".format(hostname)).json()["uuid"]
        except KeyError:
            return None

    def get_host_uuid_v2 (self, hostname):
        """
        :param hostname: str
        :return: str
        """
        logging.debug ('Reached get_host_uuid_v2')
        logging.debug ('Passing to get_host_uuid_v1')
        return self.get_host_uuid_v1 (hostname)
    
    def get_all_users(self):
        """
        :return: list
        """
        users_qty = self.get("users", params={"per_page": 1}).json()["total"]
        users = [{"firstname": user["firstname"], "lastname": user["lastname"], "login": user["login"]} for user in self.get("users", params={"per_page": users_qty}).json()["results"]]
        return users
    
    def get_all_usergroups(self):
        """
        :return: list
        """
        groups_qty = self.get("usergroups", params={"per_page": 1}).json()["total"]
        groups = [group["name"] for group in self.get("usergroups", params={"per_page": groups_qty}).json()["results"]]
        return groups
